from enum import Enum

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class AnnotationType(Enum):
    act = 1
    amount = 2
    citation = 3
    condition = 4
    constraint = 5
    copyright = 6
    court = 7
    court_citation = 8
    cusip = 9
    date = 10
    definition = 11
    distance = 12
    duration = 13
    geoentity = 14
    money = 15
    percent = 16
    pii = 17
    phone = 18
    ssn = 19
    ratio = 20
    regulation = 21
    trademark = 22
    url = 23
    laws = 24



"""Ratio extraction for English.

This module implements ratio extraction functionality in English.

Todo:
  * Improved unit tests and case coverage
"""

# Imports
import calendar
import regex as re
import string
from typing import Generator, Dict, Any

from lexnlp.extract.common.annotations.cusip_annotation import CusipAnnotation

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


CUSIP_PTN = r"""
(?:\W|^)
(?P<code>
    (?P<issuer_id>\d{2}[\dA-Z]{1}[\dA-Z]{2}[\dA-Z\@\#\*]{1})
    (?P<issue_id>[\dA-Z\@\#\*]{2})
    (?P<checksum>\d{1})
)
(?:\W|$)"""
CUSIP_PTN_RE = re.compile(CUSIP_PTN, re.VERBOSE | re.DOTALL)

INTERNAL_ISSUER_ID_PTN = r"""
(?:\d{3}99[\dA-Z]|99\d{3}[\dA-Z])
"""
INTERNAL_ISSUER_ID_PTN_RE = re.compile(INTERNAL_ISSUER_ID_PTN, re.VERBOSE)

PPN_PTN_RE = re.compile(r'[\@\#\*]')

TBA_PTN =  r"""
(?P<product_code>\d{2})
(?P<mortgage_type>[A-Z]{1})
(?P<coupon>\d{3})
(?P<maturity>\d{1})
(?P<settlement_month>\d{1})
(?P<checksum>\d{1})
"""
TBA_PTN_RE = re.compile(TBA_PTN, re.VERBOSE)
TBA_MONTHS = {i: calendar.month_name[n+1] for n, i in enumerate('123456789ABC')}

CHECKSUM_BASE = {i: 10+n for n, i in enumerate(string.ascii_uppercase)}
CHECKSUM_BASE.update({'*': 36, '@': 37, '#': 38})


def is_cusip_valid(code, return_checksum=False):
    code, _checksum = code[:8], int(code[8])
    _sum = 0
    for pos, char in enumerate(code, start=1):
        if char.isdigit():
            num = int(char)
        elif char in CHECKSUM_BASE:
            num = CHECKSUM_BASE[char]
        else:
            return False
        if pos % 2 == 0:
            num *= 2
        _sum += int(num / 10) + num % 10
    checksum = (10 - _sum % 10) % 10
    if return_checksum:
        return checksum
    return checksum == _checksum


def get_cusip(text: str) -> Generator[Dict[str, Any], None, None]:
    for ant in get_cusip_annotations(text):
        yield ant.to_dictionary_legacy()


def get_cusip_annotations(text: str) -> Generator[CusipAnnotation, None, None]:
    """
    INFO: https://www.cusip.com/pdf/CUSIP_Intro_03.14.11.pdf
    """
    for match in CUSIP_PTN_RE.finditer(text):
        capture = match.capturesdict()
        code = ''.join(capture['code'])
        issuer_id = ''.join(capture['issuer_id'])
        issue_id = ''.join(capture['issue_id'])
        checksum = int(capture['checksum'][0])
        ppn = False

        # validate CUSIP
        if not is_cusip_valid(code):
            continue

        tba = TBA_PTN_RE.fullmatch(code)
        if tba:
            tba = tba.groupdict()
            settlement_month_name = TBA_MONTHS.get(tba['settlement_month'])
            # if not settlement_month_name:
            #     continue
            tba['settlement_month_name'] = settlement_month_name
        elif PPN_PTN_RE.search(code):
            ppn = True

        internal = bool(INTERNAL_ISSUER_ID_PTN_RE.match(issuer_id))
        ant = CusipAnnotation(coords=(match.start(1), match.end(1)),
                              code=code,
                              issuer_id=issuer_id,
                              issue_id=issue_id,
                              checksum=checksum,
                              internal=internal,
                              tba=tba,
                              ppn=ppn)
        yield ant


def get_cusip_list(text):
    return list(get_cusip(text))



from typing import Tuple, List

from lexnlp.extract.common.annotations.text_annotation import TextAnnotation

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class CourtAnnotation(TextAnnotation):
    record_type = 'court'
    """
    create an object of CourtAnnotation like
    cp = CourtAnnotation(name='name', coords=(0, 100), text='text text')
    """
    def __init__(self,
                 coords: Tuple[int, int],
                 locale: str = 'en',
                 name: str = '',
                 text: str = '',
                 jurisdiction: str = '',
                 court_type: str = ''):
        super().__init__(
            locale=locale,
            coords=coords,
            name=name,
            text=text)
        self.jurisdiction = jurisdiction
        self.court_type = court_type

    def get_cite_value_parts(self) -> List[str]:
        parts = [self.name,
                 self.jurisdiction or '',
                 self.court_type or '']
        return parts

    def get_dictionary_values(self) -> dict:
        ant = dict(
            tags={
                'Extracted Entity Court Name': self.name,
                'Extracted Entity Text': self.text,
                'Extracted Entity Court Type': self.court_type,
                'Extracted Entity Court Jurisdiction': self.jurisdiction
            })
        return ant



from lexnlp.extract.common.annotations.definition_annotation import DefinitionAnnotation
from lexnlp.extract.common.annotations.text_annotation import TextAnnotation
from lexnlp.extract.common.pattern_found import PatternFound
from lexnlp.extract.common.text_pattern_collector import TextPatternCollector
from lexnlp.utils.lines_processing.line_processor import LineOrPhrase

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class UniversalDefinitionsParser(TextPatternCollector):
    """
    EsDefinitionsParser searches for definitions in text according to the
    rules of Spanish. See the "parse" method
    """

    def make_annotation_from_pattrn(self,
                                    locale: str,
                                    ptrn: PatternFound,
                                    phrase: LineOrPhrase) -> TextAnnotation:
        return DefinitionAnnotation(
            name=ptrn.name, coords=(ptrn.start, ptrn.end),
            text=phrase.text[ptrn.start: ptrn.end],
            locale=locale)

    def get_definition_dictionaries(self):
        dfs = []
        for ant in self.annotations:
            dfs.append({
                    "attrs": {
                        "start": ant.coords[0],
                        "end": ant.coords[1]
                    },
                    "tags": {
                        'Extracted Entity Type': 'definition',
                        'Extracted Entity Definition Name': ant.name,
                        'Extracted Entity Text': ant.text
                    }
                })
        return dfs



"""Universal extraction of entities for which we have full dictionaries of possible names and aliases from English text.

   Example: Courts - we have the full dictionary of known courts with their names and aliases and are able to
   search the text for each possible court.
            Geo entities - we have the full set of known geo entities and can search any text for their occurrences.


   Search methods of this module require lists of possible entities with their ids, names and sets of aliases
   in different languages.
   To allow using these methods in Celery and especially for allowing building these configuration lists once and using
   them in multiple Celery tasks it is required to allow their easy and fast serialization.
   By default Celery uses JSON serialization starting from v. 4 and does not allow serializing objects of custom
   classes out of the box. So we will have to use either dicts or tuples to avoid requiring special configuration for
   Celery. Tuples are faster.

   To avoid typos in development and utilize typization hints in IDE there are few methods in this module for operating
   tuples which represent entities and aliases. They accept named parameters lists and return tuples.
"""

import re
from typing import Union, List, Dict, Set, Tuple, Callable, Generator, Any

from lexnlp.nlp.en.tokens import get_token_list, get_stem_list

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


def entity_config(entity_id: int,
                  name: str,
                  priority: int = 0,
                  aliases: List[Union[str, Tuple]] = (),
                  name_is_alias: bool = True) -> Tuple[int, str, int, List[Tuple]]:
    """
    Create entity configuration for a possible entity with its id, name and aliases to search.
    :param entity_id: Unique identifier of the entity.
    :param name: Human-readable name to displaying in UI. Searches are made not for name but for the possible aliases -
    each one having its assigned language. And name may or may not be added to the list of search aliases.
    :param priority: Optional int priority value for the entity. Can be used for sorting. Entities with higher prio
    should be selected first.
    :param aliases: List of aliases to search for. Each alias can be either string or
    (alias, language, is_abbreviation, alias_id)
    tuple. For string - a tuple with default values is created. entity_alias() function can be used to create the alias
    tuple ensuring its components type safety in IDE.
    :param name_is_alias: If True - then add entity name to the list of aliases with undefined language.
    :return: A tuple representing the entity in format (entity_id, name, [(alias, lang, is_abbrev, alias_id), ...])
    """
    res = (entity_id, name, priority, list())

    if name_is_alias:
        add_alias_to_entity(res, name)

    if aliases:
        for alias in aliases:
            if isinstance(alias, str):
                add_alias_to_entity(res, alias)
            else:
                res[3].append(alias)
    return res


def entity_alias(alias: str, language: str = None, is_abbreviation: bool = False, alias_id: int = None) \
        -> Tuple[str, str, bool, int, str]:
    """
    Create entity alias tuple. This method is just for ensuring type safety of alias components in IDE.
    :param alias_id: Alias id. None if there is no id.
    :param alias: Alias text - 'Mississippi', 'MS', 'CAN', ...
    :param language: Language - en, de, fr, ...
    :param is_abbreviation: Is this alias representing an abbreviation or not. Abbreviations have different rules
    of searching.
    :return: A tuple representing the alias in format: (alias_text, lang, is_abbreviation, alias_id)
    """
    normalized_alias = normalize_text(alias, lowercase=not is_abbreviation)
    return alias, language, is_abbreviation, alias_id, normalized_alias


def get_entity_name(entity: Tuple[int, str, int, List[Tuple]]) -> str:
    """
    Get name of the entity.
    This method is just for more comfortable development - to avoid accessing properties of
    entities by their indexes.
    :param entity:
    :return:
    """
    return entity[1]


def get_entity_id(entity: Tuple[int, str, int, List[Tuple]]) -> int:
    """
    Get id of the entity.
    This method is just for more comfortable development - to avoid accessing properties of
    entities by their indexes.
    :param entity:
    :return:
    """
    return entity[0]


def get_entity_aliases(entity: Tuple[int, str, int, List[Tuple]]) -> List[Tuple]:
    """
    Get aliases of the entity.
    This method is just for more comfortable development - to avoid accessing properties of
    entities by their indexes.
    :param entity:
    :return:
    """
    return entity[3]


def get_entity_priority(entity: Tuple[int, str, int, List[Tuple]]) -> int:
    """
    Get priority of the entity.
    This method is just for more comfortable development - to avoid accessing properties of
    entities by their indexes.
    :param entity:
    :return:
    """
    return entity[2]


def get_alias_text(alias: Tuple[str, str, bool, int]) -> str:
    """
    Get alias text from alias tuple.
    This method is just for more comfortable development - to avoid accessing properties of
    aliases by their indexes.
    :param alias:
    :return:
    """
    return alias[0]


def get_alias_id(alias: Tuple[str, str, bool, int]) -> int:
    """
    Get alias text from alias tuple.
    This method is just for more comfortable development - to avoid accessing properties of
    aliases by their indexes.
    :param alias:
    :return:
    """
    return alias[3]


def add_alias_to_entity(entity: Tuple[int, str, int, List[Tuple]],
                        alias: str, language: str = None, is_abbreviation: bool = False, alias_id: int = None):
    """
    Add alias to entity. Entities are in the form of tuples:
    (entity_id, name, [(alias_text, lang, is_abbrev, alias_id), ...]).
    This method is just for more comfortable development - to ensure type safety and avoid accessing properties of
    entities by their indexes.
    :param entity:
    :param alias:
    :param language:
    :param is_abbreviation:
    :param alias_id: Alias id or None if identifying is not supported.
    :return:
    """
    entity[3].append(entity_alias(alias, language, is_abbreviation, alias_id))


def add_aliases_to_entity(entity: Tuple[int, str, int, List[Tuple[str, str, bool]]],
                          aliases_csv: str,
                          language: str = None,
                          is_abbreviation: bool = None,
                          alias_id: int = None,
                          csv_separator: str = ';'):
    """
    Add alias to entity. Entities are in the form of tuples:
    (entity_id, name, [(alias_text, lang, is_abbrev, alias_id), ...]).
    This method can be used if there is a comma separated list of aliases stored somewhere and they all have the same
    language and is_abbreviation value.
    This method is just for more comfortable development - to ensure type safety and avoid accessing properties of
    entities by their indexes.
    :param entity:
    :param aliases_csv:
    :param language:
    :param is_abbreviation:
    :param alias_id:
    :param csv_separator:
    :return:
    """
    if aliases_csv:
        for alias in aliases_csv.split(csv_separator):
            add_alias_to_entity(entity, alias, language, is_abbreviation, alias_id)


class SearchResultPosition:
    """
    Represents a position in the normalized source text at which one or more entities have been detected.
    One or more entities having equal aliases can be detected on a position in the text.
    """
    __slots__ = ('entities_dict', 'alias_text', 'start', 'end')

    def __init__(self,
                 entity: Tuple[int, str, int, List[Tuple]],
                 alias: Tuple[str, str, bool, int],
                 start: int,
                 end: int):
        self.entities_dict = {entity[0]: (entity, alias)}
        self.alias_text = alias[0]
        self.start = start
        self.end = end

    def __repr__(self):
        ent_str = f'{self.entities_dict}'
        ending = f'alias="{self.alias_text}", @[{self.start}, {self.end}]'
        return f'{ent_str}; {ending}'

    def add_entity(self,
                   entity: Tuple[int, str, int, List[Tuple]],
                   alias: Tuple[str, str, bool, int]):
        if entity:
            self.entities_dict[entity[0]] = (entity, alias)
        return self

    def get_entities_aliases(self):
        return list(self.entities_dict.values())

    def overlaps(self, other: 'SearchResultPosition'):
        return max(self.start, other.start) <= min(self.start + len(self.alias_text) - 1,
                                                   other.start + len(other.alias_text) - 1)


def normalize_text(text: str,
                   spaces_on_start_end: bool = True,
                   spaces_after_dots: bool = True,
                   lowercase: bool = True,
                   use_stemmer: bool = False) -> str:
    """
    Normalizes text for substring search operations - extracts tokens, joins them back with spaces,
    adds missing spaces after dots for abbreviations, e.t.c.
    Overall aim of this method is to weaken substring matching conditions by normalizing both the text
    and the substring being searched by the same way removing obsolete differences between them
    (case, punctuation, ...).
    :param text:
    :param spaces_on_start_end:
    :param spaces_after_dots:
    :param lowercase:
    :param use_stemmer: Use stemmer instead of tokenizer. When using stemmer all words will be converted to singular
    number (or to some the most plain form) before matching. When using tokenizer - the words are compared as is.
    Using tokenizer should be enough for searches for entities which exist in a single number in the real world -
    geo entities, courts, .... Stemmer is required for searching for some common objects - table, pen, developer, ...
    :return:
    """
    tokens = get_stem_list(text, lowercase=lowercase) if use_stemmer else get_token_list(text, lowercase=lowercase)
    res = ' '.join(tokens)
    if spaces_on_start_end:
        res = ' ' + res + ' '
    if spaces_after_dots:
        res = res.replace('.', ' . ').replace('  ', ' ')
    return res


def alias_is_blacklisted(alias_black_list: Union[None, Dict[str, Tuple[List[str], List[str]]]],
                         norm_alias: str,
                         alias_lang:str,
                         is_abbrev: bool) -> bool:
    if not alias_black_list:
        return False
    for lang in (alias_lang, None):
        lang_aliases = alias_black_list.get(lang)
        if lang_aliases and len(lang_aliases) >= 2:
            if norm_alias in lang_aliases[1 if is_abbrev else 0]:
                return True
    return False


def _find_entity_positions(normalized_text: str,
                           normalized_text_lowercase: str,
                           entity: Tuple[int, str, int, List[Tuple]],
                           text_languages: Union[List[str], Tuple[str], Set[str]],
                           context: Dict[int, SearchResultPosition] = None,
                           use_stemmer: bool = False,
                           abbrev_uppercase_check_range: int = 20,
                           min_alias_len: int = None,
                           alias_black_list: Union[None, Dict[str, Tuple[List[str], List[str]]]] = None):
    """
    Searches for all occurrences of name/alias of the specified entity in the specified text and fills the
    provided context dict with them.
    Operates on the provided search context - a map of alias/name positions to the SearchResultPosition entries.
    If there is a previously found name/alias at the same position in the text - the longest name/alias is
    stored in the context and the shorter one is dropped.
    So after the series of execution of this method on the shared context it will be filled with the best matching
    search results for each starting position in the text. Next these results should be ordered by start index and
    checked for intersections - to drop entries having shorter names/aliases.
    Alias languages are taken into account in this method - if a language of the source text is specified then
    only aliases of this language are being searched for.

    :param normalized_text Non-lowercase version of the normalized source text - to search for abbreviations.
    :param normalized_text_lowercase: Lowercase version of the normalized source text - to search for non-abbrevs.
    :param text_languages: If set - then only aliases of these languages will be searched for.
    :param entity:
    :param context: Map of alias/name positions in the source text to SearchResultPosition entries.
    This context can be shared between multiple executions of this functions to reach the results of the overall search
    of multiple DictEntities with the longest matching DictEntity on each position.
    Can be None - for the case of single DictEntity search.
    :param alias_black_list: Prepared black list of aliases to exclude from search.
    Should be: dict of language -> tuple (list of normalized non-abbreviations, list of normalized abbreviations)
    "None" is a key for "any" language.
    :param abbrev_uppercase_check_range: To avoid false-positives in detecting abbreviations similar to AND, OR, IN
    we need to ensure that it is not english words appeared in a piece of text written in uppercase.
    For this for each abbrev we ignore it if text[position - range : position + range] == uppercase(text[...]).
    :return:
    """

    def abbrev_in_uppercase_block(text: str, position: int, check_range:int):
        block = text[max(0, position - check_range): min(len(text), position + check_range)]
        block_upper = block.upper()
        return block == block_upper

    if context is None:
        context = dict()

    entity_aliases = get_entity_aliases(entity)
    if entity_aliases:
        for ea in entity_aliases:

            alias_text = ea[0]
            alias_lang = ea[1]
            alias_is_abbreviation = ea[2]

            # get or create normalized alias
            normalized_alias = ea[4] if len(ea) == 5 and ea[4] is not None\
                else normalize_text(alias_text, lowercase=not alias_is_abbreviation, use_stemmer=use_stemmer)

            if not alias_text or (
                            text_languages and alias_lang and alias_lang not in text_languages):
                continue
            if min_alias_len and len(alias_text) < min_alias_len:
                continue

            normalized_text_for_alias = normalized_text if alias_is_abbreviation else normalized_text_lowercase

            if alias_is_blacklisted(alias_black_list, normalized_alias, alias_lang, alias_is_abbreviation):
                continue

            start = None
            while True:
                start = normalized_text_for_alias.find(normalized_alias,
                                                       start + len(normalized_alias) - 1 if start is not None else 0)
                if start < 0:
                    break

                if alias_is_abbreviation and \
                        abbrev_in_uppercase_block(normalized_text_for_alias, start, abbrev_uppercase_check_range):
                    continue
                end = start + len(normalized_alias) - 1

                already_found = context.get(start)
                if already_found and len(already_found.alias_text) >= len(alias_text):
                    already_found.add_entity(entity, ea)
                else:
                    context[start] = SearchResultPosition(entity, ea, start, end)


class DictionaryEntity:
    def __init__(self, entity: Any, coords: Tuple[int, int]):
        self.entity = entity
        self.coords = coords

    def __repr__(self):
        ent_str = 'None'
        if self.entity:
            ent_str = str(self.entity[0])
            if len(self.entity) > 1:
                ent_str += f', {self.entity[1]}'
        coord_str = ', -'
        if self.coords:
            coord_str = f', @[{self.coords[0]}, {self.coords[1]}]'
        return ent_str + coord_str


def find_dict_entities(text: str,
                       all_possible_entities: List[Tuple[int, str, int, List[Tuple]]],
                       text_languages: Union[List[str], Tuple[str], Set[str]] = None,
                       conflict_resolving_func: Callable[[List[Tuple[int, str, List[Tuple]]]],
                                                         Tuple[List[Tuple[int, str, List[Tuple]]], Tuple]] = None,
                       use_stemmer: bool = False,
                       remove_time_am_pm: bool = True,
                       min_alias_len: int = None,
                       prepared_alias_black_list: Union[None, Dict[str, Tuple[List[str], List[str]]]] = None)\
        -> Generator[DictionaryEntity, None, None]:
    """
    Find all entities defined in the 'all_possible_entities' list appeared in the source text.
    This method takes care of leaving only the longest matching search result for the case of multiple
    entities having aliases - one being a substring of another.
    This method takes care of the language of the text and aliases - if language is specified both for the text
    and for the alias - then this alias is used only if they are the same.
    This method may detect multiple possibly matching entities at a position in the text - because there can be
    entites having the same aliases in the same language. To resolve such conflicts a special resolving function can be
    specified.
    This method takes care of time AM/PM components which possibly can appear in the aliases of some entities -
    it tries to detect minutes/seconds/milliseconds before AM/PM and ignore them in such cases.

    Algorithm of this method:
    1. Normalize the source text (we need lowercase and non-lowercase versions for abbrev searches).
    2. Create a shared search context - a map of position -> (alias text + list of matching entities)
    3. For each possible entity do search using the shared context:
        3.1. For each alias of the entity:
            3.1.1. Iteratively search for all occurrences of the alias taking into account its language, abbrev status.
                    For each found occurrence of the alias - check if there is already found another alias and entity
                    at this position and leave only the one having the longest alias ("Something" vs "Something Bigger")
                    If there is already a found different entity on this position having totally equal alias with
                    the same language - then store them both for this position in the text.
    4. Now we have a map filled with: position -> (alias text + list of entities having this alias).
    After sorting the items of this dict by position we will be able to get rid of overlaping of longer and shorter
    aliases being one a substirng of another ("Bankr. E.D.N.Y." vs "E.D.N.Y.").
    5. For each next position check if it overlaps with the next one [position; position + len(alias)].
    If overlaps - then leave the longest alias and drop the shorter.


    Main complexity of this algorithm is caused by the requirement to detect the longest match for each piece of text
    while the longer match can start at the earlier position then the shorter match and there can be multiple aliases
    of different entities matching the same piece of text.

    Another algorithm for this function can be based on the idea that or-kind regexp returns the longest matching group.
    We could form regexps containing the possible aliases and apply them to the source text:
    r'alias1|alias2|longer alias2|...'

    TODO Compare to other algorithms for time and memory complexity

    :param text:
    :param all_possible_entities: list of dict or list of DictEntity - all possible entities to search for
    :param min_alias_len: Minimal length of alias/name to search for. Can be used to ignore too short aliases like "M."
    while searching.
    :param prepared_alias_black_list: List of aliases to remove from searching. Can be used to ignore concrete aliases.
    Prepared black list of aliases to exclude from search.
    Should be: dict of language -> tuple (list of normalized non-abbreviations, list of normalized abbreviations)
    :param text_languages: If set - then only aliases of these languages will be searched for.
    :param conflict_resolving_func: A function for resolving conflicts when there are multiple entities detected
    at the same position in the source text and their detected aliases are of the same length.
    The function takes a list of conflicting entities and should return a list of one or more entities which
    should be returned.
    :param use_stemmer: Use stemmer instead of tokenizer. Stemmer converts words to their simple form (singular number,
    e.t.c.). Stemmer works better for searching for "tables", "developers", ... Tokenizer fits for "United States",
    "Mississippi", ...
    :param remove_time_am_pm: Remove from final results AM/PM abbreviations which look like end part of time
    strings - 11:45 am, 10:00 pm.
    :return:
    """

    if not text:
        return

    normalized_text = normalize_text(text, lowercase=False, use_stemmer=use_stemmer)
    normalized_text_lowercase = normalized_text.lower()

    search_context = dict()
    # Search for each DictEntity occurrence adding them into the shared search context.
    for dict_entity in all_possible_entities:
        _find_entity_positions(normalized_text, normalized_text_lowercase, dict_entity, text_languages, search_context,
                               use_stemmer=use_stemmer, min_alias_len=min_alias_len,
                               alias_black_list=prepared_alias_black_list)

    # At this moment we have a map of positions in the text
    # to SearchResultPosition entries (position + appeared name/alias + DictEntity).
    # Now we need to filter out the overlapping names/aliases - leaving only the longest one for each conflict.

    # Iterating over the (text pos -> SearchResultPosition) entries sorted by text pos.
    # For each next entry - checking if it intersects with the previous one by the means
    # of [text pos; text pos + len(found name/alias)].
    # An intersection means that we have a conflict of two courts having shorter and longer names/aliases.
    # So we leave only the longest one for each conflicting place.
    # Need to note that for the SearchResultPosition left on each step - we only need to test it for overlapping
    # with the next and further entries because they are already sorted by position in the text.
    # And this way it is done via a single loop.
    prev_pos = None

    def resolve_conflicts(pos: SearchResultPosition) \
            -> List[DictionaryEntity]:
        """
        Takes SearchResultPosition (multiple found entities+aliases at the same position in the text)
        and return a single entity+its alias which should be returned for this position or their smaller list.
        Entity is (entity_id, name, list of aliases). Alias is: (alias_text, language, is_abbrev, alias_id)
        :param pos:
        :return:
        """
        entities_at_pos = pos.get_entities_aliases()
        if remove_time_am_pm and pos.alias_text.lower() in ('am', 'pm'):
            # We should check if this is not something like 11:45 am or 11:45.123 pm
            maybe_time1 = normalized_text_lowercase[max(0, pos.start - 3):pos.start + 1]
            maybe_time2 = normalized_text_lowercase[max(0, pos.start - 4):pos.start + 1]
            if re.match(r':\d\d\s', maybe_time1) \
                    or re.match(r'\.\d\d\d\s', maybe_time2):
                return []

        if len(entities_at_pos) == 1:
            return [DictionaryEntity(entities_at_pos[0], (pos.start, pos.end))]
        else:
            cfree_ents = conflict_resolving_func(entities_at_pos) \
                if conflict_resolving_func else entities_at_pos
            return [DictionaryEntity(ent, (pos.start, pos.end))
                    for ent in cfree_ents]

    for (_index, next_pos) in sorted(search_context.items()):
        if prev_pos and not next_pos.overlaps(prev_pos):
            for entity in resolve_conflicts(prev_pos):
                yield entity
            prev_pos = next_pos
        else:
            prev_pos = prev_pos if prev_pos and len(prev_pos.alias_text) >= len(next_pos.alias_text) else next_pos

    if prev_pos:
        resolved_ents = resolve_conflicts(prev_pos)
        yield from resolved_ents


def conflicts_take_first_by_id(conflicting_entities_aliases: List[Tuple[Tuple[int, str, int, List[Tuple]], Tuple]]) \
        -> List[Tuple[Tuple[int, str, int, List[Tuple]], Tuple[str, str, bool, int]]]:
    """
    Default conflict resolving function for dropping all entities detected at the same position excepting the one
    having the smallest id. To be used in find_dict_entities() method.
    :param conflicting_entities_aliases: list of (entity, alias) pairs
    :return:
    """
    return [min(conflicting_entities_aliases, key=lambda entity_alias_pair: get_entity_id(entity_alias_pair[0])), ]


def conflicts_top_by_priority(conflicting_entities_aliases: List[Tuple[Tuple[int, str, int, List[Tuple]], Tuple]]) \
        -> List[Tuple[Tuple[int, str, int, List[Tuple]], Tuple[str, str, bool, int]]]:
    """
    Default conflict resolving function for dropping all entities detected at the same position excepting the one
    having the smallest id. To be used in find_dict_entities() method.
    :param conflicting_entities_aliases: list of (entity, alias) pairs
    :return:
    """
    return [max(conflicting_entities_aliases,
                key=lambda entity_alias_pair: get_entity_priority(entity_alias_pair[0])), ]


def prepare_alias_blacklist_dict(alias_blacklist: List[Tuple[str, str, bool]], use_stemmer: bool=False) \
        -> Union[None, Dict[str, Tuple[List[str], List[str]]]]:
    """
    Prepare alias black list for providing it to find_dict_entities() function.
    :param alias_blacklist: Non-normalized form of the blacklist: [(alias, lang, is_abbreb), ...]
    :param use_stemmer: Use stemmer for alias normalization. Otherwise - tokenizer only.
    :return:
    """
    if not alias_blacklist:
        return None
    res = dict()
    for alias, lang, is_abbrev in alias_blacklist:
        lang_tuple = res.get(lang)
        if lang_tuple is None:
            lang_tuple = ([], [])
            res[lang] = lang_tuple
        if is_abbrev:
            lang_tuple[1].append(normalize_text(alias, lowercase=False, use_stemmer=use_stemmer))
        else:
            lang_tuple[0].append(normalize_text(alias, lowercase=True, use_stemmer=use_stemmer))
    return res



import os
# pylint: disable=unused-import
from typing import List, Pattern, Generator
# pylint: enable=unused-import
import regex as re
import pandas as pd

from lexnlp.extract.common.base_path import lexnlp_base_path
from lexnlp.extract.common.annotations.regulation_annotation import RegulationAnnotation

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class RegulationsParser:
    """
    Parses Spanish regulations (acts, institutions and so on):
    - "la emisión de instrumentos inscritos en el Registro Nacional de Valores, colocados"
      boils down to 'Registro Nacional de Valores'

    - expects words like 'registro', 'comisión', 'comision', 'ley del'
      that open the following phrase
    """
    def __init__(self, regulations_dataframe: pd.DataFrame = None):
        """
        :param regulations_dataframe: a pandas dataframe with 2 columns: trigger: str, position: str
        """
        self.regulations_dataframe = regulations_dataframe
        self.start_triggers = []  # type: List[str]
        self.reg_start_triggers = []  # type: List[Pattern]
        self.load_trigger_words()
        self.setup_regexes()
        self.annotations = []  # type: List[RegulationAnnotation]
        self.locale = ''

    def setup_regexes(self) -> None:
        # read *.csv content and build regexes out of this data
        triggers_str = "|".join(self.start_triggers)
        reg = re.compile(r"(?:(?<=[\s\b])|(?<=^))(%s)[^,\b;\.\n]+" % triggers_str,
                   re.UNICODE | re.IGNORECASE)
        self.reg_start_triggers.append(reg)

    def load_trigger_words(self) -> None:
        # load records like ('ley del', 'start') - (trigger_word, position)
        dtypes = {'trigger': str,
                  'position': str}
        if not self.regulations_dataframe:
            path = os.path.join(lexnlp_base_path, 'lexnlp/config/es/es_regulations.csv')
            self.regulations_dataframe = pd.read_csv(path,
                                                     encoding="utf-8",
                                                     error_bad_lines=False,
                                                     converters=dtypes)
        subset = self.regulations_dataframe[['trigger', 'position']]
        tuples = [tuple(x) for x in subset.values]
        self.start_triggers = [t[0] for t in tuples if t[1] == 'start']

    def parse(self, text: str, locale: str = None) -> List[RegulationAnnotation]:
        # find annotations in text passed and return them as a list of objects
        self.locale = locale or 'es'
        self.annotations = []
        self.match_start_trigger(text)
        for ant in self.annotations:
            ant.country = 'Spain'
        return self.annotations

    def match_start_trigger(self, phrase: str) -> None:
        """
        :param phrase: mediante la emisión de instrumentos inscritos en el Registro Nacional de Valores, colocados
        :return: {name: 'Registro Nacional de Valores', probability: 100, ...}
        """
        for reg in self.reg_start_triggers:
            for match in reg.finditer(phrase):
                text = match.group()
                coords = (match.start(), match.end())
                ant = RegulationAnnotation(
                    name=text,
                    coords=coords,
                    text=text,
                    locale=self.locale)
                self.annotations.append(ant)

    def trim_annotations(self) -> None:
        # remove excess words from each definition
        pass

    def get_annotations_as_dictionaries(self) -> List:
        # make dictionaries like
        # { "attr": { "start": 100, "end": 162 }, "tags": {..} }
        # out of annotations
        return [a.to_dictionary() for a in self.annotations]


def make_de_regulations_parser():
    p = RegulationsParser()
    return p


parser = make_de_regulations_parser()


def get_regulation_annotations(text: str, language: str = None) -> \
        Generator[RegulationAnnotation, None, None]:
    yield from parser.parse(text, language if language else 'es')


def get_regulations(text: str, language: str = None) -> Generator[dict, None, None]:
    regs = parser.parse(text, language if language else 'es')
    for reg in regs:
        yield reg.to_dictionary()


def get_regulation_list(text: str, language: str = None) -> List[RegulationAnnotation]:
    return parser.parse(text, language if language else 'es')



from typing import Tuple, List
from lexnlp.extract.common.annotations.text_annotation import TextAnnotation

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class UrlAnnotation(TextAnnotation):
    record_type = 'url'
    """
    create an object of UrlAnnotation like
    cp = UrlAnnotation(name='name', coords=(0, 100), url='www.google.com')
    """
    def __init__(self,
                 coords: Tuple[int, int],
                 locale: str = 'en',
                 text: str = None,
                 url: str = None):
        super().__init__(
            name='',
            locale=locale,
            coords=coords,
            text=text)
        self.url = url

    def get_cite_value_parts(self) -> List[str]:
        return [self.url]

    def get_dictionary_values(self) -> dict:
        df = {
            'tags': {
                'Extracted Entity URL': self.url,
                'Extracted Entity Text': self.text
            }
        }
        return df



"""
Date extraction
Dates parser based on dateparser package
"""

# pylint: disable=bare-except,broad-except,unused-argument

import re
import pandas as pd
import string

from dateparser.search import search_dates
from typing import Generator

from lexnlp.extract.common.annotations.date_annotation import DateAnnotation
from lexnlp.extract.en.date_model import MODEL_DATE, get_date_features

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class DateParser(object):
    """
    Dates parser based on dateparser package
    """
    ENABLE_CLASSIFIER_CHECK = True
    CLASSIFIER_MODEL = MODEL_DATE
    CLASSIFIER_THRESHOLD = 0.5
    BAD_FULL_RE = re.compile(r'\d+\W?|(?:\d+\.? )?die|so|\d+ und \d+|die in einem', re.I)
    BAD_PARTIAL_RE = re.compile('[%s]' % re.escape(re.sub('[.,-:]', '', string.punctuation)))
    DATEPARSER_SETTINGS = {'PREFER_DAY_OF_MONTH': 'first', 'STRICT_PARSING': False}

    def __init__(self, text=None, language='en', dateparser_settings=None,
                 enable_classifier_check=None, classifier_model=None, classifier_threshold=None):
        """
        :param language: str - two-letters language definition
        :param enable_classifier_check: bool - enable date check using classifier model
        :param classifier_model: obj - classifier itself
        :param classifier_threshold: float 0<x<1 - min value to predict date
        :param dateparser_settings: dict - settings for dateparser
        """
        self.LANGUAGE = language
        self.TEXT = text
        self.DATES = []
        self.ENABLE_CLASSIFIER_CHECK = enable_classifier_check \
            if enable_classifier_check is not None else self.ENABLE_CLASSIFIER_CHECK
        self.CLASSIFIER_MODEL = classifier_model or self.CLASSIFIER_MODEL
        self.CLASSIFIER_THRESHOLD = classifier_threshold or self.CLASSIFIER_THRESHOLD
        self.DATEPARSER_SETTINGS = dateparser_settings or self.DATEPARSER_SETTINGS

    def get_dateparser_dates(self, text=None):
        """
        Extract possible dates with dateparser
        """
        text = text or self.TEXT
        # INFO: 'DATE_ORDER': 'DMY' prevents parsing date like 2004-12-13T00:00:00Z,
        #  use SKIP_TOKENS setting if needed along with DATE_ORDER
        return search_dates(text, languages=[self.LANGUAGE], settings=self.DATEPARSER_SETTINGS) or []

    def get_extra_dates(self):
        """
        Add custom search logic; use self.TEXT, self.LANGUAGE, self.DATES; update self.DATES
        :return: None
        """
        # self.DATES += (custom logic)

    def passed_general_check(self, date_str, date):
        """
        Apply custom checks like for unwanted symbols in a date
        """
        return not (self.BAD_FULL_RE.fullmatch(date_str) or self.BAD_PARTIAL_RE.search(date_str))

    def passed_classifier_check(self, location_start, location_end):
        """
        Use pre-trained classifier model to predict whether a date has right format
        Should be pluggable as it takes 90% parsing time
        """
        row_df = pd.DataFrame.from_records(
            [get_date_features(self.TEXT, location_start, location_end)])
        date_score = self.CLASSIFIER_MODEL.predict_proba(
            row_df.loc[:, self.CLASSIFIER_MODEL.columns])
        return date_score[0, 1] > self.CLASSIFIER_THRESHOLD

    def get_dates(self, text=None, language=None):
        for ant in self.get_date_annotations(text, language):
            yield {'location_start': ant.coords[0],
                   'location_end': ant.coords[1],
                   'value': ant.date,
                   'source': ant.text}

    def get_date_annotations(self,
                             text: str = None,
                             language: str = None) -> \
            Generator[DateAnnotation, None, None]:

        self.TEXT = text.replace('\n', ' ') or self.TEXT
        self.LANGUAGE = language or self.LANGUAGE

        if not self.TEXT or not self.LANGUAGE:
            raise RuntimeError('Define text and language.')

        # First try dateparser searcher
        try:
            self.DATES = self.get_dateparser_dates() or []
        except Exception as e:
            # TODO: add logging
            print(str(e))

        # Next try custom search logic
        self.get_extra_dates()

        positions = []
        for date_str, date in sorted(self.DATES, key=lambda i: -len(i[0])):

            # if possible date has weird format or unwanted symbols
            if not self.passed_general_check(date_str, date):
                continue

            for match in re.finditer(re.escape(date_str), self.TEXT):
                location_start, location_end = match.span()

                # skip overlapping entities
                if any([1 for i, j in positions if location_start>=i and location_end<=j]):
                    continue
                positions.append(match.span())

                # filter out possible dates using classifier
                if self.ENABLE_CLASSIFIER_CHECK and \
                        not self.passed_classifier_check(location_start, location_end):
                    continue

                ant = DateAnnotation(coords=(location_start, location_end),
                                     date=date,
                                     text=self.TEXT[location_start:location_end],
                                     locale=language or self.LANGUAGE)
                yield ant


    def get_date_list(self, *args, **kwargs):
        return list(self.get_dates(*args, **kwargs))


get_dates = DateParser().get_dates
get_date_list = DateParser().get_date_list



import regex as re
from typing import List, Tuple, Generator
from lexnlp.extract.common import year_parser
from lexnlp.extract.common.annotations.court_citation_annotation import CourtCitationAnnotation
from lexnlp.extract.de.dates import get_dates
from lexnlp.utils.lines_processing.phrase_finder import PhraseFinder

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class PossibleToken:
    def __init__(self, token_type: str, value: str, coords: Tuple[int, int], prob: int):
        self.token_type = token_type
        self.value = value
        self.coords = coords
        self.prob = prob

    def __repr__(self):
        return '%s [%s] at (%d, %d), prob: %d' % \
               (self.value, self.token_type, self.coords[0], self.coords[1], self.prob)


class CourtCitationsParser:
    """
    https://docs.google.com/spreadsheets/d/1_-Hnr46s8JmTYIFSkcI01gbwqwjnVDoAV1dUv3Bz5fM/edit#gid=0

    could be one of the following:

    (lalala; DATE; vom) or (lalala; DATE; Beschluss)

    or § lalala Court Name (abbreviation)

    or
    """
    reg_cite_chunk = re.compile(r"\([^0-9\)]+[0-9\.]{2,}[^\)]+\)", re.UNICODE)
    reg_trigger_words = re.compile('Beschluss|vom', re.UNICODE | re.IGNORECASE)
    reg_token_end = re.compile(r"[;,]")

    registries = {
        'BStBl': 'Bundessteuerblatt (sonstige gebräuchliche Verwendung)',
        'BFH': 'Deutschland, Rechtswesen: Bundesfinanzhof',
        'BFHE': 'Sammlung der Entscheidungen des BFH',
        'GmS-OGB': 'Beschluss des Gemeinsamen Senats der obersten Gerichtshöfe des Bundes',
        'BVerwGE': 'Entscheidungen des Bundesverwaltungsgerichts',
        'GrS': 'Beschluss des Großen Senats des BFH',
        'BFH-Urteile': 'Bundesfinanzhof Urteile',
        'BFH-Beschlüsse': 'Bundesfinanzhof Beschlüsse',
        'DstR': 'Deutsches Steuerrecht - DStR',
        'KStG':'Körperschaftsteuergesetz'
    }

    registry_finder = None
    reg_split_by_registry = None

    # region STATICINIT
    if not registry_finder:
        registry_finder = PhraseFinder(list(registries.keys()))
    if not reg_split_by_registry:
        reg_split_by_registry = re.compile("|".join(list(registries.keys())))
    # endregion

    def __init__(self):
        self.locale = 'de'
        self.items = []  # List[CourtCitationAnnotation]
        self.locale = None

    def parse(self, text: str, locale: str = None) -> List[CourtCitationAnnotation]:
        self.items = []
        self.locale = locale
        self.find_citations_in_embraced_text(text)
        return self.items

    def find_citations_in_embraced_text(self, text: str) -> None:
        fragment_start = 0
        for embraced_text in CourtCitationsParser.reg_cite_chunk.finditer(text):
            start = embraced_text.start()

            # process text before braces
            fragment = text[fragment_start:start]
            self.split_chunk_and_find_citations(fragment, fragment_start)
            fragment_start = embraced_text.end() + 1

            # process text in braces
            self.process_chunks_in_embraced_text(embraced_text, start)

        fragment = text[fragment_start:-1]
        self.split_chunk_and_find_citations(fragment, fragment_start)

    def process_chunks_in_embraced_text(self, embraced_text: str, start) -> None:
        parts = embraced_text.group().split(';')
        for part in parts:
            self.get_detail_from_chunk(part, start)
            start += len(part) + 1
        return

    def split_chunk_and_find_citations(self, text: str, start: int) -> None:
        chunks = self.split_text_by_keywords(text)
        for chunk in chunks:
            self.get_detail_from_chunk(chunk[0], chunk[1] + start)

    def get_detail_from_chunk(self, chunk_text: str, chunk_start: int) -> None:
        chunk_body = chunk_text.strip(r'() \t')
        dates = self.get_dates_from_text(chunk_body)
        registries = self.get_registries_from_text(chunk_body)
        triggers = CourtCitationsParser.reg_trigger_words.search(chunk_body)

        if not triggers and len(registries) == 0 and len(dates) == 0:
            return

        start = chunk_start + chunk_text.find(chunk_body)
        end = start + len(chunk_body)
        ant = CourtCitationAnnotation(name=chunk_body,
                                      coords=(start, end),
                                      text=chunk_body,
                                      locale=self.locale)
        ant.locale = self.locale
        if len(registries) > 0:
            ant.name = CourtCitationsParser.registries[registries[0].value]
            ant.short_name = self.get_reference_from_registry(registries[0], chunk_body)
        self.items.append(ant)

    def get_reference_from_registry(self, registry: PossibleToken,
                                    chunk_body: str) -> str:
        start = registry.coords[0]
        end = -1
        end_match = CourtCitationsParser.reg_token_end.search(chunk_body[start:])
        if end_match:
            end = end_match.start()
        return chunk_body[start: end + 1].strip(' \t.,;()')

    def get_registries_from_text(self, text: str) -> List[PossibleToken]:
        reg_names = [(m, 100) for m in CourtCitationsParser.registry_finder.find_word(text, ignore_case=False)]
        # if the case is not the same, the probability is 50%
        reg_names += [(m, 50) for m in CourtCitationsParser.registry_finder.find_word(text, ignore_case=True)]
        reg_names.sort(key=lambda n: n[0][1] - n[1] * 1000)

        toks = []
        for match_prob in reg_names:
            tok = PossibleToken('registry', match_prob[0][0],
                                (match_prob[0][1], match_prob[0][2]),
                                match_prob[1])
            toks.append(tok)
        return toks

    def get_dates_from_text(self, text: str) -> List[PossibleToken]:
        try:
            date_ents = list(get_dates(text))
        except TypeError:
            date_ents = []
        date_ents.sort(key=lambda d: d['location_start'])

        tokens = []
        for d in date_ents:
            tok = PossibleToken('date', d['value'],
                                (d['location_start'], d['location_end']), 100)
            tokens.append(tok)
        if len(tokens) > 0:
            return tokens

        # try get years only
        for year in year_parser.year_parser.get_years_with_coords_from_string(text):
            tokens.append(PossibleToken('date', str(year[0]),
                                        (year[1], year[2]), 50))
        return tokens

    def split_text_by_keywords(self, text: str) -> List[Tuple[str, int]]:
        matches =  list(CourtCitationsParser.reg_split_by_registry.finditer(text))
        chunks = []
        for i in range(len(matches)):
            match = matches[i]
            ending = -1
            if i < len(matches) - 1:
                ending = matches[i + 1].start() - 1
            phrase_break = CourtCitationsParser.reg_token_end.search(text[match.end():ending])
            if phrase_break:
                ending = min(ending, phrase_break.start())
            chunks.append((text[match.start():ending], match.start()))
        return chunks


parser = CourtCitationsParser()


def get_court_citation_annotations(text: str, language: str = None) -> \
        Generator[CourtCitationAnnotation, None, None]:
    yield from parser.parse(text, language if language else 'de')


def get_court_citations(text: str, language: str = None) -> Generator[dict, None, None]:
    cts = parser.parse(text, language if language else 'de')
    for ct in cts:
        yield ct.to_dictionary()


def get_court_citation_list(text: str, language: str = None) -> List[CourtCitationAnnotation]:
    return parser.parse(text, language if language else 'de')



"""Utility methods for segmentation classifiers

This module implements utility methods for segmentation, such as shared methods to generate
document character distributions.

Todo:
"""

# Imports
import string

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


def build_document_distribution(text, characters=string.printable, norm=True):
    """
    Build document character distribution based on fixed character, optionally norming.
    :param text:
    :param characters:
    :param norm:
    :return:
    """
    # Check for empty
    text_len = len(text)
    if text_len == 0:
        return {}

    # Build character vector
    char_vector = {}

    for character in characters:
        char_vector["doc_char_{0}".format(character)] = text.count(character)

    # Norm if requested
    if norm:
        total = float(sum(char_vector.values()))
        for key in char_vector:
            char_vector[key] = char_vector[key] / total

    return char_vector


def build_document_line_distribution(text, characters=string.printable, norm=True):
    """
    Build document and line character distribution for section segmenting based on fixed character, optionally
    normalizing vector.
    """

    # Build character vector
    feature_vector = {}
    for character in characters:
        feature_vector["doc_char_{0}".format(character)] = text.count(character)
        feature_vector["doc_startchar_{0}".format(character)] = 0
    feature_vector["doc_startchar_other"] = 0

    # Build line start vector
    for line in text.splitlines():
        if len(line.strip()) > 0:
            character = line.strip()[0]

            if character in characters:
                feature_vector["doc_startchar_{0}".format(character)] += 1
            else:
                feature_vector["doc_startchar_other"] += 1
        else:
            continue

    # Norm if requested
    if norm:
        total_char = float(sum([b for a, b in feature_vector.items() if a.startswith("doc_char")]))
        total_startchar = float(sum([b for a, b in feature_vector.items() if a.startswith("doc_startchar")]))

        for character in feature_vector.keys():
            if character.startswith("doc_char"):
                feature_vector[character] = feature_vector[character] / total_char
            elif character.startswith("doc_startchar"):
                feature_vector[character] = feature_vector[character] / total_startchar if total_startchar != 0 else 0

    return feature_vector



"""Extraction utilities for English.
"""

# Imports
import re
import string
import unicodedata
from itertools import groupby

import nltk
from typing import Generator, List, Tuple

from lexnlp.extract.common.annotations.phrase_position_finder import PhrasePositionFinder

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


# Default punctuation to whitelist
VALID_PUNCTUATION = [".", ",", "'", "-", "&", "(", ")"]


def strip_unicode_punctuation(text, valid_punctuation=None):
    """
    This method strips all unicode punctuation that is not whitelisted.
    :param text: text to strip
    :param valid_punctuation: valid punctuation to whitelist
    :return:
    """
    valid_punctuation = valid_punctuation or VALID_PUNCTUATION

    return "".join(c for c in text if (c in valid_punctuation) or not unicodedata.category(c).startswith("P"))


default_grammar = r"""
    NBAR:
        {<DT>?<NNP.*|JJ|POS|\(|\)|,>*<NNP.*>}  # DeTerminer, Proper Noun, Adjective, brackets, terminated by Proper Noun
    IN:
        {<CC|IN>}   # Coordinating Conjunction, Preposition/Subordinating Conjunction
    NP:
        {(<NBAR><IN>)*<NBAR>}
"""


class NPExtractor(object):

    exception_sym = ['&', 'and', 'of']
    exception_pos = ['IN', 'CC']
    sym_with_space = ['(', '&']
    sym_without_space = [i for i in string.punctuation if i not in ['(', '&']] + ["'s"]

    def __init__(self, grammar=None):
        grammar = grammar or default_grammar
        self.chunker = nltk.RegexpParser(grammar)

    def get_tokenizer(self):
        tokenizer = nltk.tokenize.TreebankWordTokenizer
        tokenizer.PUNCTUATION[4] = (re.compile(r'[;@#$%]', re.UNICODE), ' \\g<0> ')
        return tokenizer().tokenize

    def cleanup_leaves(self, leaves):
        leaves = [l for l in
                  [list(group) for key, group in groupby(
                      leaves, key=lambda k: k[0] not in self.exception_sym and k[1] in self.exception_pos)]
                  if l[0][1] not in self.exception_pos or l[0][0] in self.exception_sym]
        return leaves

    def get_np(self, text: str) -> Generator[str, None, None]:
        tokenizer_func = self.get_tokenizer()
        tokens = tokenizer_func(text)
        pos_tokens = nltk.tag.pos_tag(tokens)
        chunks = self.chunker.parse(pos_tokens)

        for tree in chunks.subtrees(filter=lambda t: t.label() == 'NP'):
            leaves = self.cleanup_leaves(tree.leaves())
            for np_items in leaves:
                yield self.join(np_items)

    def get_np_with_coords(self, text: str) -> List[Tuple[str, int, int]]:
        phrases = list(self.get_np(text))
        tagged_phrases = PhrasePositionFinder.find_phrase_in_source_text(
            text, phrases)
        return tagged_phrases

    def join(self, np_items):
        np = ''
        last_pos = None
        for n, current_pos in enumerate(np_items):
            item, _ = current_pos
            sep = self.sep(n, current_pos, last_pos)
            np += sep + item
            last_pos = current_pos
        return self.strip_np(np)

    def sep(self, n, current_pos, last_pos):
        return ' ' if n and current_pos[0] not in self.sym_without_space\
                      and last_pos[0] not in "(" else ""

    @staticmethod
    def strip_np(np):
        return np.strip(string.punctuation.replace(')', '') + string.whitespace)



"""
Addresses extraction for English language.
"""

import os
import pickle
import re
from typing import Generator, Tuple, List

from lexnlp.extract.en.preprocessing.span_tokenizer import SpanTokenizer

from lexnlp.extract.en.addresses import address_features

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


NGRAM_CLASSIFIER_FN = os.path.join(os.path.dirname(__file__), 'addresses_clf.pickle')


class Address:
    def __init__(self, zip_code: str,
                 country: str,
                 state: str,
                 city: str,
                 addr1: str,
                 addr2: str) -> None:
        super().__init__()
        self.zip_code = zip_code
        self.country = country
        self.state = state
        self.city = city
        self.addr1 = addr1
        self.addr2 = addr2 if addr2 != addr1 else None

    def __str__(self) -> str:
        return '{0}, {1}, {2}, {3}, {4}, {5}'.format(self.addr1, self.addr2, self.city, self.state,
                                                     self.country, self.zip_code)

    def members(self):
        return self.zip_code, self.country, self.state, self.city, self.addr1, self.addr2

    def __eq__(self, other):
        if type(other) is type(self):
            return self.members() == other.members()
        else:
            return False

    def __hash__(self):
        return hash(self.members())


class NGramType:
    OTHER = 0
    ADDR_START = 1
    ADDR_MIDDLE = 2
    ADDR_END = 3


TOKENIZER = SpanTokenizer()


def _safe_index(sentence, token, point, safe: bool = False):
    try:
        return sentence.index(token, point)
    except ValueError:
        if safe:
            return None
        else:
            raise ValueError(f'Substring "{token}" not found in:\n'
                             f'"{sentence}"\n'
                             f'Search start pos: {point}')


def align_tokens(tokens, sentence):
    """
    Copy of the same function from nltk fixing processing of double quotes.
    :param tokens:
    :param sentence:
    :return:
    """
    point = 0
    offsets = []
    for token in tokens:
        if token == '``' or token == "''":
            start = _safe_index(sentence, '"', point, True)
            if start is None:
                start = _safe_index(sentence, '\'', point)
            point += 1
        else:
            start = _safe_index(sentence, token, point)
            point = start + len(token)
        offsets.append((start, point))
    return offsets


def load_classifier():
    with open(NGRAM_CLASSIFIER_FN, 'rb') as f:
        return pickle.load(f)


NGRAM_CLASSIFIER = load_classifier()
NGRAM_WINDOW_HALF_WIDTH = 10
NGRAM_WINDOW_STEP = 1


def prepare_ngrams_in_text(text: str, window_half_width: int, window_step: int) \
        -> Generator[Tuple[List[int], List[str], int, int], None, None]:
    words2 = []

    for word, pos_token, word_start_pos, word_end_pos in TOKENIZER.get_token_spans(text):
        features = address_features.get_word_features(word, pos_token)
        # our tokenizer returns exact word_end_pos and we need it so that text[word_start_pos:word_end_pos] == word
        words2.append((word, pos_token, word_start_pos, word_end_pos + 1, features))

    i = 0
    while i < len(words2):
        word, pos_token, word_start_pos, word_end_pos, features = words2[i]
        features = list()
        for j in range(i - window_half_width, i + window_half_width):
            if 0 <= j < len(words2):
                features.extend(words2[j][4])
            else:
                features.extend(address_features.ZERO_FEATURES)
        yield features, word, word_start_pos, word_end_pos
        i += window_step


_MARGIN_TOLERANCE = 2


def cleanup(address: str) -> str:
    address = address.strip('?:!.,;-_ \t')
    address = re.sub(r'\s+', ' ', address)
    return address


def get_addresses(text: str) -> Generator[str, None, None]:
    for addr, _start, _end in get_address_spans(text):
        yield addr


def get_address_spans(text: str) -> Generator[Tuple[str, int, int], None, None]:
    possible_address_start = None
    possible_address_end = None
    margin = 0
    for ngram_features, _word, word_start_pos, word_end_pos \
            in prepare_ngrams_in_text(text, NGRAM_WINDOW_HALF_WIDTH, NGRAM_WINDOW_STEP):
        ngram_type = NGRAM_CLASSIFIER.predict([ngram_features])

        if possible_address_start is None:
            if ngram_type in (NGramType.ADDR_START, NGramType.ADDR_MIDDLE):
                possible_address_start = word_start_pos
                possible_address_end = word_end_pos
                margin = 0
        else:
            if ngram_type in (NGramType.ADDR_MIDDLE, NGramType.ADDR_END):
                possible_address_end = word_end_pos
                margin = 0
            elif ngram_type == NGramType.OTHER:
                if margin >= _MARGIN_TOLERANCE:
                    if possible_address_end - possible_address_start >= 20:
                        possible_address = text[possible_address_start:possible_address_end]
                        yield cleanup(possible_address), possible_address_start, possible_address_end
                    possible_address_end = None
                    possible_address_start = None
                    margin = 0
                else:
                    margin += 1
            else:  # ngram start again
                possible_address_end = word_end_pos
                possible_address_start = word_start_pos
                margin = 0

    if possible_address_start is not None:  # text ends with address
        possible_address = text[possible_address_start:possible_address_end]
        yield cleanup(possible_address), possible_address_start, possible_address_end



"""Date extraction for English.

This module implements date extraction functionality in English.
"""

# pylint: disable=bare-except

# Standard imports
import itertools
import os
import string

from sklearn.externals import joblib

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


# Setup path
MODULE_PATH = os.path.dirname(os.path.abspath(__file__))

# Load model
MODEL_DATE = joblib.load(os.path.join(MODULE_PATH, "./date_model.pickle"))

DATE_MODEL_CHARS = []
DATE_MODEL_CHARS.extend(string.ascii_letters)
DATE_MODEL_CHARS.extend(string.digits)
DATE_MODEL_CHARS.extend(["-", "/", " ", "%", "#", "$"])


def get_date_features(text, start_index, end_index, include_bigrams=True, window=5, characters=None,
                      norm=True):
    """
    Get features to use for classification of date as false positive.
    :param text: raw text around potential date
    :param start_index: date start index
    :param end_index: date end index
    :param include_bigrams: whether to include bigram/bicharacter features
    :param window: window around match
    :param characters: characters to use for feature generation, e.g., digits only, alpha only
    :param norm: whether to norm, i.e., transform to proportion
    :return:
    """
    # Check chars
    if not characters:
        characters = DATE_MODEL_CHARS

    # Get text window
    window_start = max(0, start_index - window)
    window_end = min(len(text), end_index + window)
    feature_text = text[window_start:window_end].strip()

    # Build character vector
    char_vec = {}
    char_keys = []
    bigram_keys = {}
    for character in characters:
        key = "char_{0}".format(character)
        char_vec[key] = feature_text.count(character)
        char_keys.append(key)

    # Build character bigram vector
    if include_bigrams:
        bigram_set = ["".join(s) for s in itertools.permutations(characters, 2)]
        bigram_keys = []
        for character in bigram_set:
            key = "bigram_{0}".format(character)
            char_vec[key] = feature_text.count(character)
            bigram_keys.append(key)

    # Norm if requested
    if norm:
        # Norm by characters
        char_sum = sum([char_vec[k] for k in char_keys])
        if char_sum > 0:
            for key in char_keys:
                char_vec[key] /= float(char_sum)

        # Norm by bigrams
        if include_bigrams:
            bigram_sum = sum([char_vec[k] for k in bigram_keys])
            if bigram_sum > 0:
                for key in bigram_keys:
                    char_vec[key] /= float(bigram_sum)

    return char_vec



__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class PatternFound:
    """
    used inside EsDefinitionsParser and SpanishParsingMethods
    to store intermediate parsing results
    """
    def __init__(self):
        self.name = None # type: str
        self.start = 0
        self.end = 0
        self.probability = 0

    # pylint: disable=unused-argument
    def pattern_worse_than_target(self, p, text: str) -> bool:  # p: PatternFound
        """
        check what pattern is better then 2 patterns are considered duplicated
        "text" may be used in derived classes
        """
        spans = self.start <= p.start <= self.end and \
                self.start <= p.end <= self.end
        if not spans:
            return False
        return self.name.find(p.name) >= 0



"""Copyright extraction for English using NLTK and NLTK pre-trained maximum entropy classifier.

This module implements basic Copyright extraction functionality in English relying on the pre-trained
NLTK functionality, including POS tagger and NE (fuzzy) chunkers.
"""

# Imports
import regex as re
import string
from typing import Generator, List, Tuple

from lexnlp.extract.common.annotations.copyright_annotation import CopyrightAnnotation

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class CopyrightEnStyleParser:
    reg_company_name = re.compile(r"[\p{Lu}]+[\p{L}\s]*", re.UNICODE)

    reg_valid_company_name = re.compile(r"\p{L}[\p{L}\s,]+", re.UNICODE)

    year_ptn = r"(\d{4}(?:\s*[-,–]\s*\d{4})?)"
    year_ptn_re = re.compile(year_ptn + '$')

    copyright_ptn = fr"((Copyright\W\s*|\(\s*[Cc]\s*\)\s*|©)+\s*{year_ptn}?\s*(.+))"
    copyright_ptn_re = re.compile(copyright_ptn)

    copyright_dates_re = re.compile(r'\d{2,}')

    @staticmethod
    def get_copyright(text: str,
                      return_sources=False) -> Generator[CopyrightAnnotation, None, None]:
        for ant in CopyrightEnStyleParser.get_copyright_annotations(
                text, return_sources):
            ret = (ant.sign,
                   ant.date,
                   ant.name)
            if return_sources:
                ret += (ant.text,)
            yield ret

    @classmethod
    def extract_phrases_with_coords(cls, sentence: str) -> List[Tuple[str, int, int]]:
        raise NotImplementedError()

    @classmethod
    def get_copyright_annotations(cls, text: str, return_sources=False) \
            -> Generator[CopyrightAnnotation, None, None]:
        """
        Find copyright in text.
        :param text:
        :param return_sources:
        :return:
        """
        # Iterate through sentences
        if not cls.copyright_ptn_re.search(text):
            return

        tagged_phrases = cls.extract_phrases_with_coords(text)

        for phrase, phrase_start, phrase_end in tagged_phrases:
            for match in cls.copyright_ptn_re.finditer(phrase):
                cp_text, cp_sign, cp_date, cp_name = match.groups()

                # TODO: catch in the general regex
                if not cp_date:
                    cp_date_at_end = cls.year_ptn_re.search(cp_name)
                    if cp_date_at_end:
                        cp_date = cp_date_at_end.group()
                        cp_name = re.sub(r'{}$'.format(cp_date), '', cp_name)

                start, end = match.span()
                if end > (phrase_end - phrase_start):
                    end = phrase_end - phrase_start
                start += phrase_start
                end += phrase_start
                ant = CopyrightAnnotation(coords=(start, end),
                                          sign=cp_sign.strip(),
                                          date=cp_date,
                                          name=cp_name.strip(string.punctuation +
                                                             string.whitespace))

                if return_sources:
                    ant.text = cp_text.strip()
                cls.split_copyright_date(ant)
                cls.derive_company_name(ant, phrase)
                yield ant

    @classmethod
    def derive_company_name(cls,
                           ant: CopyrightAnnotation,
                           phrase: str) -> None:
        if ant.company:
            ant.company = ant.company.strip(' ,;-(:')
            if cls.reg_valid_company_name.search(ant.company):
                return
            ant.company = ''
        possible_names = [n.group(0) for n in cls.reg_company_name.finditer(ant.name)]
        if not possible_names:
            possible_names = [n.group(0) for n in cls.reg_company_name.finditer(phrase)]
        if possible_names:
            ant.company = cls.take_best_company_name(possible_names)
            ant.company = ant.company.strip(' ,;-(:')

    @classmethod
    def take_best_company_name(cls, names: List[str]) -> str:
        # e.g., ['Huawei', 'permiten que más']
        for name in names:
            if cls.reg_company_name.search(name):
                return name
        return names[0]

    @classmethod
    def split_copyright_date(cls, ant: CopyrightAnnotation) -> None:
        if not ant.date:
            return
        years = [int(y.group()) for y in cls.copyright_dates_re.finditer(ant.date)]
        if len(years) == 2:
            if 10000 > years[0] > 100 and years[1] >= years[0]:
                ant.year_start = years[0]
                ant.year_end = years[1]
                return

        if len(years) == 1 and 10000 > years[0] > 100:
            ant.year_start = years[0]



"""Amount extraction for English.

This module implements basic amount extraction functionality in English.

This module supports converting:
- numbers with comma delimiter: "25,000.00", "123,456,000"
- written numbers: "Seven Hundred Eighty"
- mixed written numbers: "5 million" or "2.55 BILLION"
- written ordinal numbers: "twenty-fifth"
- fractions (non written): "1/33", "25/100"; where 1 < numerator < 99; 1 < denominator < 999
- fraction No/100 wil be treated as 00/100
- written numbers and fractions: "twenty one AND 5/100"
- written fractions: "one-third", "three tenths", "ten ninety-ninths", "twenty AND one-hundredths",
  "2 hundred and one-thousandth";
   where 1 < numerator < 99 and 2 < denominator < 99 and numerator < denominator;
   or 1 < numerator < 99 and denominator == 100, i.e. 1/99 - 99/100;
   or 1 < numerator < 99 and denominator == 1000, i.e. 1/1000 - 99/1000;
- floats starting with "." (dot): ".5 million"
- "dozen": "twenty-two DOZEN"
- "half": "Six and a HALF Billion", "two and a half"
- "quarter": "five and one-quarter", "5 and one-quarter", "three-quartes"
- multiple numbers: "$25,400, 1 million people and 3.5 tons"

Avoids:
- skip: "5.3.1.", "1/1/2010"
"""

# pylint: disable=bare-except

# Imports
import string
from typing import Generator

import nltk
import regex as re
from num2words import num2words

from lexnlp.extract.common.annotations.amount_annotation import AmountAnnotation

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


# Define small numbers
SMALL_NUMBERS = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
                 20, 30, 40, 50, 60, 70, 80, 90]
SMALL_NUMBERS_MAP = {num2words(n): n for n in SMALL_NUMBERS}
SMALL_NUMBERS_MAP.update({num2words(n, ordinal=True): n for n in SMALL_NUMBERS})
SMALL_NUMBERS_MAP.update({num2words(n, ordinal=True) + 's': n for n in SMALL_NUMBERS[3:20]})
SMALL_NUMBERS_MAP.update({num2words(n).replace('y', 'ieths'): n for n in SMALL_NUMBERS[20:]})

BIG_NUMBERS_EXPONENT = [3, 6, 9, 12]
MAGNITUDE_MAP = {num2words(10 ** n)[4:]: 10 ** n for n in BIG_NUMBERS_EXPONENT}
MAGNITUDE_MAP.update(
    {'thousandth': 1000,
     'thousandths': 1000,
     'trill': 1000000000000,
     'bil': 1000000000,
     'mm': 1000000,
     'k': 1000,
     'm': 1000000,
     'b': 1000000000})

small_numbers = list(SMALL_NUMBERS_MAP.keys())
small_numbers.sort(key=len, reverse=True)
big_numbers = list(MAGNITUDE_MAP.keys())
big_numbers.sort(key=len, reverse=True)

CURRENCY_SYMBOL_MAP = {
    "$": "USD",
    "€": "EUR",
    "¥": "JPY",
    "£": "GBP",
    "₠": "EUR",
    "₨": "INR",
    "₹": "INR",
    "₺": "TRY",
    "元": "CNY",
    "₽": "RUB",
    # "¢": None,
    "₩": "KRW",
}
CURRENCY_PREFIX_MAP = {
    "chf": "CHF",
    "rmb": "CNY",
}
allowed_prev_units = list(CURRENCY_SYMBOL_MAP) + list(CURRENCY_PREFIX_MAP)

fraction_smb_to_value = {
    '½': 1.0/2, '⅓': 1.0/3, '⅔': 2.0/3,
    '¼': 1.0/4, '¾': 3.0/4, '⅕': 1.0/5,
    '⅖': 2.0/5, '⅗': 3.0/5, '⅘': 4.0/5,
    '⅙': 1.0/6, '⅚': 5.0/6, '⅐': 1.0/7,
    '⅛': 1.0/8, '⅜': 3.0/8, '⅝': 5.0/8,
    '⅞': 7.0/8, '⅑': 1.0/9, '⅒': 1.0/10
}
fraction_smb_to_string = {k: str(fraction_smb_to_value[k])[1:]
                          for k in fraction_smb_to_value}

fraction_symbols = ''.join([k for k in fraction_smb_to_value])
FRACTION_TAIL = rf'\s{{0,2}}[{fraction_symbols}]+'
FRACTION_TAIL_RE = re.compile(FRACTION_TAIL,
                              re.IGNORECASE | re.MULTILINE | re.DOTALL | re.VERBOSE)

NUM_PTN = r"""
(?:(?:(?:(?:(?:[\.\d][\d\.,]*\s*|\W|^)
(?:(?:{written_small_numbers}|{written_big_numbers}
|hundred(?:th(?:s)?)?|dozen|and|a\s+half|quarters?)[\s-]*)+)
(?:(?:no|\d{{1,2}})/100)?)|(?<=\W|^)(?:[\.\d][\d\.,/]*))(?:\W|$))(?:{fraction_tail})*""".format(
    written_small_numbers='|'.join(small_numbers),
    written_big_numbers='|'.join(big_numbers),
    fraction_tail=FRACTION_TAIL)

NUM_PTN_RE = re.compile(NUM_PTN, re.IGNORECASE | re.MULTILINE | re.DOTALL | re.VERBOSE)

NON_WRIT_RE = re.compile(r'[\d\.]+')
MIXED_WRIT_RE = re.compile(r'(^[\d\.]*)(.+)', re.DOTALL)
ONLY_BIG_WRIT_RE = re.compile(r'^\s*(?:{}|hundred|dozen)'.format('|'.join(MAGNITUDE_MAP)))
NUM_FRACTION_RE = re.compile(r'(\s+no|\d{1,2})/(\d{1,3}[^/])')
NUM_FRACTION_SUB_RE = re.compile(r'(?:\s*and)?(?:\s+no|\s*\d{1,2})/\d{1,3}')
HALF_RE = re.compile(r'\s*and\s+a\s+half')
QUARTER_RE = re.compile(r'(?:\s*and\s+)?(one|two|three)[\s-]+quarters?')
AND_RE = re.compile(r'\W*and\W*', re.IGNORECASE | re.MULTILINE | re.DOTALL)

FRACTION_PTN = r"(?:(?:\W|^)" \
               r"(?:one[\s-]+(?:{writ_ord_2_90}|hundredth|thousandth|(?:{writ_20_90})[\s-]+" \
               r"(?:{writ_ord_1_9})))|" \
               r"(?:(?:{writ_1_90}|[\s-]+)+[\s-]+(?:{writ_ord_3_90_pl}|hundredths|thousandths" \
               r"(?:{writ_20_90})[\s-]+(?:{writ_ord_1_9_pl}))))" \
               r"(?:\W|$)".format(writ_ord_2_90='|'.join([num2words(n, ordinal=True) for n in SMALL_NUMBERS[2:]]),
                                  writ_20_90='|'.join([num2words(n) for n in SMALL_NUMBERS[20:]]),
                                  writ_ord_1_9='|'.join([num2words(n, ordinal=True) for n in SMALL_NUMBERS[1:10]]),
                                  writ_1_90='|'.join([num2words(n) for n in SMALL_NUMBERS[1:]]),
                                  writ_ord_3_90_pl='|'.join([num2words(n, ordinal=True) + 's'
                                                             for n in SMALL_NUMBERS[3:]]),
                                  writ_ord_1_9_pl='|'.join([num2words(n, ordinal=True) + 's'
                                                            for n in SMALL_NUMBERS[1:10]]))
FRACTION_PTN_RE = re.compile(FRACTION_PTN)

FRACTION_EXTRACT_PTN = r"((?:(?:{writ})|(?:(?:{writ_20_90})[\s-]+(?:{writ_1_9}))))[\s-]+" \
                       r"((?:{writ_ord_mix}|(?:(?:{writ_20_90})[\s-]+" \
                       r"(?:{writ_ord_1_9_mix}))|hundredths?|thousandths?))" \
    .format(writ='|'.join([num2words(n) for n in SMALL_NUMBERS[1:]]),
            writ_20_90='|'.join([num2words(n) for n in SMALL_NUMBERS[20:]]),
            writ_1_9='|'.join([num2words(n) for n in SMALL_NUMBERS[1:10]]),
            writ_ord_mix='|'.join([num2words(n, ordinal=True) + 's?' for n in SMALL_NUMBERS[2:]]),
            writ_ord_1_9_mix='|'.join([num2words(n, ordinal=True) + 's?' for n in SMALL_NUMBERS[1:10]]))
FRACTION_EXTRACT_PTN_RE = re.compile(FRACTION_EXTRACT_PTN, re.S | re.M)

wnl = nltk.stem.WordNetLemmatizer()

# Taken from Su Nam Kim Paper...
grammar = r"""
    NBAR:
        {<NN.*|JJ>*<NN.*>}  # Nouns and Adjectives, terminated with Nouns

    NP:
        {<NBAR>}
        {<NBAR><IN><NBAR>}  # Above, connected with in/of/etc...
"""
chunker = nltk.RegexpParser(grammar)


def text2num(s, search_fraction=True):
    """
    Convert written amount into integer/float.
    :param s: written number
    :param search_fraction: extract fraction
    :return: integer/float
    """
    n = 0
    g = 0
    s = s.lower().replace(',', '').replace('-', ' ').strip(string.whitespace).rstrip(
        string.punctuation + string.whitespace)
    s = re.sub(r'\s+and\s*$|^\s*and\s+', '', s)
    if not (s.startswith('.') and s[1].isdigit()):
        s = s.lstrip(string.punctuation + string.whitespace)
    if s in ['k', 'm', 'b']:
        return
    # if only number or float in string
    if NON_WRIT_RE.fullmatch(s):
        return float(s)

    # if written number has integer/float prefix: "25 million", "2.035 thousand tons"
    if not NUM_FRACTION_RE.fullmatch(s):
        p, s = MIXED_WRIT_RE.search(s).groups()
        g = float(p) if p else 0

    # if written big number has no prefix: "lovely million", "a dozen"
    if ONLY_BIG_WRIT_RE.search(s) and not g:
        s = 'one ' + s

    d = 0
    dnd = NUM_FRACTION_RE.search(s)
    fs = FRACTION_PTN_RE.search(s)
    q = QUARTER_RE.search(s)
    if q:  # convert quarters
        s = QUARTER_RE.sub('', s)
        nu = q.groups()[0]
        d = text2num(nu) / 4
    elif dnd:  # if text has fraction like 1/33 or 87/100 or 1/100
        dn, dd = dnd.groups()
        if dn.isdigit():
            d = int(dn) / int(dd)
        s = NUM_FRACTION_SUB_RE.sub('', s)
    elif fs and search_fraction:  # extract written fractions
        try:
            s = FRACTION_PTN_RE.sub('', s)
            fe = fs.group(0)
            fn, fd = FRACTION_EXTRACT_PTN_RE.search(fe).groups()
            fn = text2num(fn, search_fraction=False)
            fd = text2num(fd, search_fraction=False)
            d = fn / fd
        except (ValueError, TypeError, ZeroDivisionError):
            pass

    # process
    a = s.split()

    x1 = 0
    for w in a:
        if w in ['a', 'and']:
            continue
        x = SMALL_NUMBERS_MAP.get(w, None)
        if x is not None:
            g += x
        elif 'hundred' in w and g != 0:
            g *= 100
        elif w == 'dozen' and g != 0:
            g *= 12
        elif w == 'half':
            if x1:
                g += x1 * .5
            else:
                g += .5
        else:
            x = x1 = MAGNITUDE_MAP.get(w, None)
            if x is not None:
                n += g * x
                g = 0
            else:
                raise RuntimeError('Unknown number: ' + w)
    return n + g + d


def get_np(text) -> Generator:
    tokens = nltk.word_tokenize(text)
    pos_tokens = nltk.tag.pos_tag(tokens)
    chunks = chunker.parse(pos_tokens)
    for subtree in chunks.subtrees(filter=lambda t: t.label() == 'NP'):
        np = ' '.join([i[0] for i in subtree.leaves()])
        _np = ' '.join([wnl.lemmatize(i[0]) for i in subtree.leaves()])
        yield np, _np


def get_amounts(text: str,
                return_sources=False,
                extended_sources=True,
                float_digits=4) -> Generator[float, None, None]:
    """
    Find possible amount references in the text.
    :param text: text
    :param return_sources: return amount AND source text
    :param extended_sources: return data around amount itself
    :param float_digits: round float to N digits, don't round if None
    :return: list of amounts
    """
    for ant in get_amount_annotations(text, extended_sources, float_digits):  # type: AmountAnnotation
        if return_sources:
            yield (ant.value, ant.text)
        else:
            yield ant.value


def get_amount_annotations(text: str,
                           extended_sources=True,
                           float_digits=4) \
        -> Generator[AmountAnnotation, None, None]:
    """
    Find possible amount references in the text.
    :param text: text
    :param extended_sources: return data around amount itself
    :param float_digits: round float to N digits, don't round if None
    :return: list of amounts
    """
    for match in NUM_PTN_RE.finditer(text):
        found_item = match.group()
        fract_tail_items = FRACTION_TAIL_RE.finditer(found_item)
        for fract_tail in fract_tail_items:
            fract_tail_smb = fract_tail.group().strip(' ')
            if fract_tail_smb in fraction_smb_to_string:
                fract_ending = fraction_smb_to_string[fract_tail_smb]
                found_item = found_item[:fract_tail.span()[0]]
                found_item += fract_ending
            break

        if AND_RE.fullmatch(found_item):
            continue
        try:
            amount = text2num(found_item)
        except:
            continue
        if amount is None:
            continue
        if isinstance(amount, float) and float_digits:
            amount = round(amount, float_digits)

        if extended_sources:
            unit = ''
            next_text = text[match.span()[1]:]
            if next_text:
                for np, _ in get_np(next_text):
                    if next_text.startswith(np):
                        unit = np
                if unit:
                    found_item = ' '.join([found_item.strip(), unit])
            if not unit:
                prev_text = text[:match.span()[0]]
                prev_text_tags = nltk.word_tokenize(prev_text)
                if prev_text_tags and prev_text_tags[-1].lower() in allowed_prev_units:
                    sep = ' ' if text[match.span()[0] - 1] == ' ' else ''
                    found_item = sep.join([prev_text_tags[-1], found_item.rstrip()])

            ant = AmountAnnotation(coords=match.span(),
                                   value=amount,
                                   text=found_item.strip())
            yield ant
        else:
            ant = AmountAnnotation(coords=match.span(),
                                   value=amount,
                                   text=match.group())
            yield ant



from typing import Generator, List, Tuple

from lexnlp.extract.common.annotation_locator_type import AnnotationLocatorType
from lexnlp.extract.common.annotations.definition_annotation import DefinitionAnnotation
from lexnlp.extract.en.definition_parsing_methods import DefinitionCaught, get_definition_list_in_sentence, \
    filter_definitions_for_self_repeating
from lexnlp.extract.ml.en.definitions.layered_definition_detector import LayeredDefinitionDetector
from lexnlp.nlp.en.segments.sentences import get_sentence_span

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


def get_definitions_in_sentence(sentence: str,
                                return_sources=False,
                                decode_unicode=True) -> Generator:
    definitions = get_definition_list_in_sentence((0, len(sentence), sentence), decode_unicode)
    for df in definitions:
        if return_sources:
            yield df.name, df.text
        else:
            yield df.name


def get_definition_objects_list(text, decode_unicode=True) -> List[DefinitionCaught]:
    """
    :param text: text to search for definitions
    :param decode_unicode:
    :return: a list of found definitions - objects of class DefinitionCaught
    """
    definitions = []
    for sentence in get_sentence_span(text):  # type: Tuple[int, int, str]
        definitions += get_definition_list_in_sentence(sentence, decode_unicode)
    definitions = filter_definitions_for_self_repeating(definitions)
    return definitions


parser_ml_classifier = LayeredDefinitionDetector()


def get_definition_annotations(text: str,
                               decode_unicode=True,
                               locator_type: AnnotationLocatorType = AnnotationLocatorType.RegexpBased) \
        -> Generator[DefinitionAnnotation, None, None]:

    if locator_type == AnnotationLocatorType.MlWordVectorBased:
        if not parser_ml_classifier.initialized:
            raise Exception(f'"parser_ml_classifier" object should be initialized (call load_compressed method)')
        ants = parser_ml_classifier.get_annotations(text)
        for ant in ants:
            yield ant
        return

    # use Regexp-based locator
    for d in get_definition_objects_list(text,
                                         decode_unicode=decode_unicode):
        ant = DefinitionAnnotation(coords=d.coords,
                                   text=d.text,
                                   name=d.name)
        yield ant


def get_definitions(text: str,
                    return_sources=False,
                    decode_unicode=True,
                    return_coords=False,
                    locator_type: AnnotationLocatorType = AnnotationLocatorType.RegexpBased) -> Generator:
    """
    Find possible definitions in natural language in text.
    The text will be split to sentences first.
    :param return_coords: returns a (x, y) tuple in each record. x - definition's text start, y - definition's text end
    :param decode_unicode:
    :param return_sources: returns a tuple with the extracted term and the source sentence
    :param text: the input text
    :param locator_type: use default (Regexp-based) or ML-based locator
    :return: Generator[name] or Generator[name, text] or Generator[name, text, coords]
    """
    if locator_type == AnnotationLocatorType.MlWordVectorBased:
        if not parser_ml_classifier.initialized:
            raise Exception(f'"parser_ml_classifier" object should be initialized (call load_compressed method)')
        definitions = parser_ml_classifier.get_annotations(text)
    else:
        definitions = get_definition_objects_list(text, decode_unicode)

    for df in definitions:
        if return_coords:
            yield df.name, df.text, (df.coords[0], df.coords[1])
        elif return_sources:
            yield df.name, df.text
        else:
            yield df.name


def get_definitions_explicit(text,
                             decode_unicode=True,
                             locator_type: AnnotationLocatorType = AnnotationLocatorType.RegexpBased) -> Generator:
    yield from get_definitions(text,
                               return_sources=True,
                               decode_unicode=decode_unicode,
                               return_coords=True,
                               locator_type=locator_type)



from typing import Tuple, List, Optional
from lexnlp.utils.map import Map
from lexnlp.extract.common.annotations.text_annotation import TextAnnotation

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class CompanyAnnotation(TextAnnotation):
    record_type = 'company'
    """
    create an object of CompanyAnnotation like
    cp = CompanyAnnotation(name='name', coords=(0, 100), text='text text')
    """
    def __init__(self,
                 coords: Tuple[int, int],
                 locale: str = 'en',
                 name: Optional[str] = None,
                 name_abbr: Optional[str] = None,
                 company_type_full: Optional[str] = None,
                 company_type_abbr: Optional[str] = None,
                 company_type_label: Optional[str] = None,
                 description: Optional[str] = None,
                 text: Optional[str] = None,
                 counter: int = 1):
        super().__init__(
            name=name,
            coords=coords,
            locale=locale)
        self.name_abbr = name_abbr
        self.company_type_full = company_type_full
        self.company_type_abbr = company_type_abbr
        self.company_type_label = company_type_label
        self.description = description
        self.text = text
        self.counter = counter

    @property
    def company_type(self) -> str:
        return self.company_type_full or self.company_type_abbr \
               or self.company_type_label or ''

    def __repr__(self):
        text = self.name or self.name_abbr or self.text or ''
        if self.company_type:
            text += f' {self.company_type}'
        return f'{text}, ({self.coords[0]}, {self.coords[1]})'

    def get_cite_value_parts(self) -> List[str]:
        parts = [self.name, str(self.company_type)]
        return parts

    def get_dictionary_values(self) -> dict:
        df = Map({
            'tags': {
                'Extracted Entity Name': self.name,
                'Extracted Entity Text': self.text or self.name
            }
        })
        if self.name:
            df.tags["Extracted Entity Company"] = self.name
        if self.company_type:
            df.tags["Extracted Entity Company Type"] = self.company_type
        return df



"""Geo Entity extraction for English.

This module implements extraction functionality for geo entities in English, including formal names, abbreviations,
and aliases.

"""

from typing import List, Tuple, Union, Dict, Generator, Any

from lexnlp.extract.common.annotations.text_annotation import TextAnnotation
from lexnlp.extract.common.annotations.geo_annotation import GeoAnnotation
from lexnlp.config.en import geoentities_config
from lexnlp.extract.en.dict_entities import find_dict_entities, conflicts_take_first_by_id, \
    prepare_alias_blacklist_dict, conflicts_top_by_priority, entity_config, add_aliases_to_entity

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


_ALIAS_BLACK_LIST_PREPARED = prepare_alias_blacklist_dict(geoentities_config.ALIAS_BLACK_LIST)


def get_geoentities(text: str,
                    geo_config_list: List[Tuple[int, str, List[Tuple[str, str, bool, int]]]],
                    priority: bool = False,
                    priority_by_id: bool = False,
                    text_languages: List[str] = None,
                    min_alias_len: int = geoentities_config.MIN_ALIAS_LEN,
                    prepared_alias_black_list: Union[None, Dict[str, Tuple[List[str], List[str]]]]
                    = _ALIAS_BLACK_LIST_PREPARED) -> Generator[Tuple[Tuple, Tuple], Any, Any]:
    """
    Searches for geo entities from the provided config list and yields pairs of (entity, alias).
    Entity is: (entity_id, name, [list of aliases])
    Alias is: (alias_text, lang, is_abbrev, alias_id)

    This method uses general searching routines for dictionary entities from dict_entities.py module.
    Methods of dict_entities module can be used for comfortable creating the config: entity_config(),
    entity_alias(), add_aliases_to_entity().
    :param text:
    :param geo_config_list: List of all possible known geo entities in the form of tuples
    (id, name, [(alias, lang, is_abbrev, alias_id), ...]).
    :param priority: If two entities found with the totally equal matching aliases -
    then use the one with the greatest priority field.
    :param priority_by_id: If two entities found with the totally equal matching aliases -
    then use the one with the lowest id.
    :param text_languages: Language(s) of the source text. If a language is specified then only aliases of this
    language will be searched for. For example: this allows ignoring "Island" - a German language
     alias of Iceland for English texts.
    :param min_alias_len: Minimal length of geo entity aliases to search for.
    :param prepared_alias_black_list: List of aliases to exclude from searching in the form:
     dict of lang -> (list of normalized non-abbreviation aliases, list of normalized abbreviation aliases).
     Use dict_entities.prepare_alias_blacklist_dict() for preparing this dict.
    :return: Generates tuples: (entity, alias)
    """
    conflict_resolving_func = None

    if priority_by_id:
        conflict_resolving_func = conflicts_take_first_by_id

    if priority:
        conflict_resolving_func = conflicts_top_by_priority

    for ent in find_dict_entities(text,
                                  geo_config_list,
                                  conflict_resolving_func=conflict_resolving_func,
                                  text_languages=text_languages,
                                  min_alias_len=min_alias_len,
                                  prepared_alias_black_list=prepared_alias_black_list):
        yield ent.entity


def get_geoentity_annotations(text: str,
                    geo_config_list: List[Tuple[int, str, List[Tuple[str, str, bool, int]]]],
                    priority: bool = False,
                    priority_by_id: bool = False,
                    text_languages: List[str] = None,
                    min_alias_len: int = geoentities_config.MIN_ALIAS_LEN,
                    prepared_alias_black_list: Union[None, Dict[str, Tuple[List[str], List[str]]]]
                    = _ALIAS_BLACK_LIST_PREPARED) -> Generator[GeoAnnotation, None, None]:
    "See get_geoentities"

    conflict_resolving_func = None

    if priority_by_id:
        conflict_resolving_func = conflicts_take_first_by_id

    if priority:
        conflict_resolving_func = conflicts_top_by_priority

    dic_entries = find_dict_entities(text,
                                     geo_config_list,
                                     conflict_resolving_func=conflict_resolving_func,
                                     text_languages=text_languages,
                                     min_alias_len=min_alias_len,
                                     prepared_alias_black_list=prepared_alias_black_list)

    for ent in dic_entries:
        ant = GeoAnnotation(coords=ent.coords)
        if ent.entity[0]:
            toponim = ent.entity[0]
            year = TextAnnotation.get_int_value(toponim[0])
            if year:
                ant.year = year
            ant.name = toponim[1]
        yield ant


def load_entities_dict_by_path(entities_fn: str, aliases_fn: str):
    entities = {}
    import csv

    with open(entities_fn, 'r', encoding='utf8') as f:
        reader = csv.DictReader(f)
        for row in reader:
            entities[row['id']] = entity_config(row['id'], row['name'], int(row['priority']) if row['priority'] else 0,
                                                name_is_alias=True)

    with open(aliases_fn, 'r', encoding='utf8') as f:
        reader = csv.DictReader(f)
        for row in reader:
            entity = entities.get(row['entity_id'])
            if entity:
                add_aliases_to_entity(entity,
                                      row['alias'],
                                      row['locale'],
                                      row['type'].startswith('iso') or row['type'] == 'abbreviation')

    return entities.values()



"""Company types and abbreviations.

This module defines constants and configuration for company types, including their full descriptions, common short
names, and abbreviations.

Todo:
  * Include automated generation from raw source.
  * Expand full description list
"""

import csv
import os

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


_COMPANY_TYPES = [
    'A.G.', 'AG', 'B.V.', 'C.A.', 'C.V.', 'Corp.', 'Corporation',
    'G.P.', 'Inc.', 'Incorporated', 'K.K.', 'L.L.C.', 'L.L.L.P.', 'L.L.P.', 'L.P.',
    'LLC.', 'LLP.', 'Lda.', 'Ltd.', 'Ltda.', 'N.A.', 'National Association',
    'N.V.', 'Pte. Ltd.',
    'S de R.L. de C.V.', 'S. DE R.L. DE C.V.', 'S. De R.L. De C.V.',
    'S. De R.L. de C.V.', 'S. de R. L. de C.V.', 'S. de R.L. de C.V.',
    'S.A.', 'S.A. DE C.V.', 'S.A. De C.V.', 'S.A. de C.V.', 'S.A.R.L.',
    'S.A.S.', 'S.A.U.', 'S.A.de C.V.', 'S.C.', 'S.C.A.',
    'S.L.', 'S.L.U.', 'S.R.L.', 'S.R.L. de C.V.', 'S.R.O.', 'S.a.r.l.',
    'S.de R.L. de C.V.', 'S.r.l.', 'Sdn. Bhd.', 's.r.o.', 'CO'
]


default_company_types_file_path = os.path.join(os.path.dirname(__file__), 'company_types.csv')


def get_company_types(file_path=None):
    ret = dict()
    file_path = file_path or default_company_types_file_path
    with open(file_path, encoding="utf-8") as f:
        for row in csv.DictReader(f):
            alias_dot = row['Alias'].strip(' ').lower()
            alias = alias_dot.strip('.')
            abbr = row['Abbreviation'].strip()
            label = row['Label'].strip()
            ret[alias] = dict(
                abbr=abbr,
                label=label)
            if alias_dot != alias:
                ret[alias_dot] = dict(
                    abbr=abbr,
                    label=label)
    return ret


COMPANY_TYPES = get_company_types()

COMPANY_DESCRIPTIONS = ['Trust Bank', 'Trust Company', 'Trust',
                        'Bank', 'Company', 'Partnership', 'Agency']



from typing import Tuple, List
from lexnlp.utils.map import Map
from lexnlp.extract.common.annotations.text_annotation import TextAnnotation

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class RegulationAnnotation(TextAnnotation):
    record_type = 'regulation'
    """
    create an object of RegulationAnnotation like
    cp = RegulationAnnotation(name='name', coords=(0, 100), text='text text')
    """
    def __init__(self,
                 coords: Tuple[int, int],
                 locale: str = 'en',
                 name: str = '',
                 text: str = None,
                 source: str = '',
                 country: str = ''):
        super().__init__(
            name=name,
            locale=locale,
            coords=coords,
            text=text)
        self.country = country
        self.source = source

    def get_cite_value_parts(self) -> List[str]:
        parts = [self.country or '',
                 self.source or '',
                 self.name or '']
        return parts

    def get_dictionary_values(self) -> dict:
        dic = Map({
            'tags': {
                'External Reference Issuing Country': self.country,
                'External Reference Text': self.name,
                'Extracted Entity Text': self.text or self.name
            }
        })
        if self.source:
            dic.tags['External Reference Source'] = self.source
        return dic

    def to_dictionary_legacy(self) -> dict:
        return {"regulation_type": self.source,
                "regulation_code": self.name,
                "regulation_text": self.text}



from lexnlp.extract.common.dates import DateParser

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


parser = DateParser(enable_classifier_check=False, language='de',
                    dateparser_settings={'PREFER_DAY_OF_MONTH': 'first',
                                         'STRICT_PARSING': False,
                                         'DATE_ORDER': 'DMY'})

get_date_annotations = parser.get_date_annotations

get_dates = parser.get_dates

get_date_list = parser.get_date_list



"""
Detect whether document is a contract.
"""

import os
import pickle

# Third-party imports
import gensim.models.doc2vec
from sklearn.externals import joblib

# LexNLP
from lexnlp.nlp.en.segments.sentences import get_sentence_list
from lexnlp.nlp.en.tokens import get_stem_list

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


data_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data')

# Load doc2vec model
d2v_model_filename = "d2v_all_size100_window10.model"
d2v_model_path = os.path.join(data_dir, d2v_model_filename)

# Load doc2vec model from part files like d2v_all_size100_window10.model.part.aa
if os.path.exists(d2v_model_path):
    d2v_model = gensim.models.doc2vec.Doc2Vec.load(d2v_model_path)
else:
    d2v_model_filenames = sorted([i for i in os.listdir(data_dir)
                                  if i.startswith('{}.part.'.format(d2v_model_filename))])
    if not d2v_model_filenames:
        raise RuntimeError('Doc2Vec model file "{}" not found'.format(d2v_model_filename))
    d2v_model_pickled = b''
    for filename in d2v_model_filenames:
        d2v_model_pickled += open(os.path.join(data_dir, filename), 'rb').read()
    d2v_model = pickle.loads(d2v_model_pickled)


# Load classifier model
rf_model = joblib.load(
    os.path.join(data_dir, "is_contract_classifier.pickle"))


# Utility methods
def process_sentence(sentence):
    return [s for s in get_stem_list(sentence, stopword=True, lowercase=True) if s.isalpha()]


def process_document(document):
    doc_words = []
    for sentence in get_sentence_list(document):
        doc_words.extend(process_sentence(sentence))
    return doc_words


def is_contract(text, min_probability=0.5, return_probability=False):
    # Create the vector representation from doc2vec model
    text_vector = d2v_model.infer_vector(process_document(text))

    # Pass vector into classifier
    try:
        classifier_score = rf_model.predict_proba([text_vector])[0, 1]
    except IndexError:
        return None

    ret = bool(classifier_score >= min_probability)

    if return_probability:
        ret = (ret, classifier_score)

    return ret



"""Ratio extraction for English.

This module implements ratio extraction functionality in English.

Todo:
  * Improved unit tests and case coverage
"""

# Imports
import regex as re
from typing import Generator

from lexnlp.extract.common.annotations.ratio_annotation import RatioAnnotation
from lexnlp.extract.en.amounts import get_amounts, NUM_PTN

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


RATIO_PTN = r"""
(({num_ptn_1})\s*
(?:to|\:|/)\s*
({num_ptn_2}))(?!\s*[ap].?m(?:\W|$))
""".format(
    num_ptn_1=NUM_PTN.replace('(?:(?:no|\\d{1,2})/100)?', '').replace('(?:\\W|$)', ''),
    num_ptn_2=NUM_PTN.replace('(?:(?:no|\\d{1,2})/100)?', '')
)
RATIO_PTN_RE = re.compile(RATIO_PTN, re.IGNORECASE | re.MULTILINE | re.DOTALL | re.VERBOSE)


def get_ratios(text: str, return_sources=False, float_digits=4) \
    -> Generator:
    for ant in get_ratio_annotations(text, float_digits=float_digits):
        item = (ant.left, ant.right, ant.ratio)
        if return_sources:
            item += (ant.text,)
        yield item


def get_ratio_annotations(text: str, float_digits=4) \
        -> Generator[RatioAnnotation, None, None]:
    for match in RATIO_PTN_RE.finditer(text.lower()):
        source_text, ratio_1_text, ratio_2_text = match.groups()
        amount_1 = list(get_amounts(ratio_1_text, float_digits=float_digits))
        amount_2 = list(get_amounts(ratio_2_text, float_digits=float_digits))
        if len(amount_1) != 1 or len(amount_2) != 1:
            continue
        amount_1 = amount_1[0]
        amount_2 = amount_2[0]
        if amount_1 == 0 or amount_2 == 0:
            continue
        if float_digits:
            amount_1 = round(amount_1, float_digits)
            amount_2 = round(amount_2, float_digits)
        total = float(amount_1) / amount_2
        ant = RatioAnnotation(coords=match.span(),
                              text=source_text.strip(),
                              left=amount_1,
                              right=amount_2,
                              ratio=total)
        yield ant



import pickle

import pandas
import os

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


_MODULE_PATH = os.path.dirname(os.path.abspath(__file__))
_FN_UNICODE_CHAR_CATEGORIES = os.path.join(_MODULE_PATH, 'unicode_character_categories.pickle')
_FN_UNICODE_CHAR_CATEGORY_MAPPING = os.path.join(_MODULE_PATH, 'unicode_character_category_mapping.pickle')
_FN_UNICODE_CHAR_TOP_CATEGORY_MAPPING = os.path.join(_MODULE_PATH, 'unicode_character_top_category_mapping.pickle')


def _load_table(fn: str, ignore_error: bool):
    try:
        with open(fn, 'rb') as f:
            return pickle.load(f)
    except IOError as e:
        print('Unable to load unicode lookup table: {0}'.format(fn))
        if ignore_error:
            return None
        else:
            raise e


UNICODE_CHAR_CATEGORIES = _load_table(_FN_UNICODE_CHAR_CATEGORIES, True)
UNICODE_CHAR_CATEGORY_MAPPING = _load_table(_FN_UNICODE_CHAR_CATEGORY_MAPPING, True)
UNICODE_CHAR_TOP_CATEGORY_MAPPING = _load_table(_FN_UNICODE_CHAR_TOP_CATEGORY_MAPPING, True)


def build_lookup_tables(fn_char_categories,
                        fn_char_category_mapping,
                        fn_char_top_category_mapping,
                        table_source: str = None):
    """
    https://www.unicode.org/reports/tr44/#General_Category
    https://www.unicode.org/reports/tr44/#General_Category_Values

    Table 12. General_Category Values

    Abbr	Long	Description
    Lu	Uppercase_Letter	an uppercase letter
    Ll	Lowercase_Letter	a lowercase letter
    Lt	Titlecase_Letter	a digraphic character, with first part uppercase
    LC	Cased_Letter	Lu | Ll | Lt
    Lm	Modifier_Letter	a modifier letter
    Lo	Other_Letter	other letters, including syllables and ideographs
    L	Letter	Lu | Ll | Lt | Lm | Lo
    Mn	Nonspacing_Mark	a nonspacing combining mark (zero advance width)
    Mc	Spacing_Mark	a spacing combining mark (positive advance width)
    Me	Enclosing_Mark	an enclosing combining mark
    M	Mark	Mn | Mc | Me
    Nd	Decimal_Number	a decimal digit
    Nl	Letter_Number	a letterlike numeric character
    No	Other_Number	a numeric character of other type
    N	Number	Nd | Nl | No
    Pc	Connector_Punctuation	a connecting punctuation mark, like a tie
    Pd	Dash_Punctuation	a dash or hyphen punctuation mark
    Ps	Open_Punctuation	an opening punctuation mark (of a pair)
    Pe	Close_Punctuation	a closing punctuation mark (of a pair)
    Pi	Initial_Punctuation	an initial quotation mark
    Pf	Final_Punctuation	a final quotation mark
    Po	Other_Punctuation	a punctuation mark of other type
    P	Punctuation	Pc | Pd | Ps | Pe | Pi | Pf | Po
    Sm	Math_Symbol	a symbol of mathematical use
    Sc	Currency_Symbol	a currency sign
    Sk	Modifier_Symbol	a non-letterlike modifier symbol
    So	Other_Symbol	a symbol of other type
    S	Symbol	Sm | Sc | Sk | So
    Zs	Space_Separator	a space character (of various non-zero widths)
    Zl	Line_Separator	U+2028 LINE SEPARATOR only
    Zp	Paragraph_Separator	U+2029 PARAGRAPH SEPARATOR only
    Z	Separator	Zs | Zl | Zp
    Cc	Control	a C0 or C1 control code
    Cf	Format	a format control character
    Cs	Surrogate	a surrogate code point
    Co	Private_Use	a private-use character
    Cn	Unassigned	a reserved unassigned code point or a noncharacter
    C	Other	Cc | Cf | Cs | Co | Cn
    """
    # https://www.unicode.org/reports/tr44/#UnicodeData.txt

    url = table_source or 'ftp://ftp.unicode.org/Public/11.0.0/ucd/UnicodeData.txt'
    df = pandas.read_table(url, sep=';', header=None)

    df.columns = ["value", "name", "general_category", "canonical_combining_class",
                  "bidi_class", "decomposition_type", "decomposition_mapping",
                  "numeric_type", "numeric_value", "bidi_mirrored",
                  "unicode_1_name", "iso_comment", "simple_uppercase_mapping",
                  "simple_lowercase_mapping", "simple_titlecase_mapping", ]

    # Get all punctuation characters
    punctuation_category_list = ["Pc", "Pd", "Ps", "Pe", "Pi", "Pf", "Po", "P"]
    punctuation_character_set = df.loc[df["general_category"].isin(punctuation_category_list), "value"].apply(
        lambda x: chr(int(x, 16))).tolist()
    print(punctuation_character_set[0:10])

    # Get all punctuation characters
    punctuation_start_category_list = ["Ps", "Pi"]
    punctuation_start_character_set = df.loc[
        df["general_category"].isin(punctuation_start_category_list), "value"].apply(lambda x: chr(int(x, 16))).tolist()
    print(punctuation_start_character_set[0:10])

    # Get all punctuation characters
    punctuation_end_category_list = ["Pe", "Pf"]
    punctuation_end_character_set = df.loc[df["general_category"].isin(punctuation_end_category_list), "value"].apply(
        lambda x: chr(int(x, 16))).tolist()
    print(punctuation_end_character_set[0:10])

    # Get all punctuation characters
    symbol_currency_category_list = ["Sc"]
    symbol_currency_character_set = df.loc[df["general_category"].isin(symbol_currency_category_list), "value"].apply(
        lambda x: chr(int(x, 16))).tolist()
    print(symbol_currency_character_set[0:10])

    # Get all math characters
    symbol_math_category_list = ["Sm"]
    symbol_math_character_set = df.loc[df["general_category"].isin(symbol_math_category_list), "value"].apply(
        lambda x: chr(int(x, 16))).tolist()
    print(symbol_math_character_set[0:10])

    # Get all math characters
    symbol_category_list = ["Sk", "So"]
    symbol_character_set = df.loc[df["general_category"].isin(symbol_category_list), "value"].apply(
        lambda x: chr(int(x, 16))).tolist()
    # symbol_character_set[0:10]

    # Get all math characters
    whitespace_category_list = ["Zs", "Zl", "Zp", "Cc", "Cf", "Cs"]
    whitespace_character_set = df.loc[df["general_category"].isin(whitespace_category_list), "value"].apply(
        lambda x: chr(int(x, 16))).tolist()
    print(whitespace_character_set[0:10])

    # Get all math characters
    space_category_list = ["Zs"]
    space_character_set = df.loc[df["general_category"].isin(space_category_list), "value"].apply(
        lambda x: chr(int(x, 16))).tolist()
    print(space_character_set[0:10])

    # Get all math characters
    line_category_list = ["Zl"]
    line_character_set = df.loc[df["general_category"].isin(line_category_list), "value"].apply(
        lambda x: chr(int(x, 16))).tolist()
    print(line_character_set[0:10])

    # Get all math characters
    para_category_list = ["Zp"]
    para_character_set = df.loc[df["general_category"].isin(para_category_list), "value"].apply(
        lambda x: chr(int(x, 16))).tolist()
    print(para_character_set[0:10])

    unicode_character_categories = {"punctuation": punctuation_character_set,
                                    "punctuation_start": punctuation_start_character_set,
                                    "punctuation_end": punctuation_end_character_set,
                                    "symbol": symbol_character_set,
                                    "symbol_currency": symbol_currency_character_set,
                                    "symbol_math": symbol_math_character_set,
                                    "whitespace": whitespace_character_set,
                                    "space": space_character_set,
                                    "line": line_character_set
                                    }

    with open(fn_char_categories, 'wb') as output_file:
        pickle.dump(unicode_character_categories, output_file)

    # Reverse mapping
    unicode_character_category_mapping = {}

    for _, row in df.iterrows():
        unicode_character_category_mapping[chr(int(row['value'], 16))] = row['general_category']

    with open(fn_char_category_mapping, 'wb') as output_file:
        pickle.dump(unicode_character_category_mapping, output_file)

    # Reverse mapping
    unicode_character_top_category_mapping = {}

    for _, row in df.iterrows():
        unicode_character_top_category_mapping[chr(int(row['value'], 16))] = row['general_category'][0].upper()

    with open(fn_char_top_category_mapping, 'wb') as output_file:
        pickle.dump(unicode_character_top_category_mapping, output_file)


if __name__ == '__main__':
    print('Building and saving unicode lookup tables...')
    build_lookup_tables(_FN_UNICODE_CHAR_CATEGORIES, _FN_UNICODE_CHAR_CATEGORY_MAPPING,
                        _FN_UNICODE_CHAR_TOP_CATEGORY_MAPPING)

    print('Done')



import copy
import logging
import regex as re
from dateutil import tz, parser
from typing import Tuple, List, Dict

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


logger = logging.getLogger('datefinder')


class DateFragment:
    '''
    This class describes big chunks of text that may contain date strings
    Each chunk includes of one of more tokens
    Each token is build upon DATE_REGEX matches
    '''

    def __init__(self):
        self.match_str = ''
        self.indices = (0, 0)
        self.captures = {}  # type:Dict[str, List[str]]
        self.matches_count = 0

    def __repr__(self):
        str_capt = ', '.join(['"{}": [{}]'.format(c, self.captures[c]) for c in self.captures])
        return '{} [{}, {}]\nCaptures: {}'.format(self.match_str, self.indices[0], self.indices[1], str_capt)

    def get_captures_count(self):
        return sum([len(self.captures[m]) for m in self.captures])

    def get_captures_group_count(self):
        return sum([1 if len(self.captures[m]) > 0 else 0 for m in self.captures])


class DateFinder(object):
    """
    Locates dates in a text
    """

    DIGITS_MODIFIER_PATTERN = r'\d+st|\d+th|\d+rd|first|second|third|fourth|fifth|sixth|seventh|eighth|nineth|tenth|next|last'
    DIGITS_PATTERN = r'\d{1,4}'
    DAYS_PATTERN = 'monday|tuesday|wednesday|thursday|friday|saturday|sunday|mon|tue|tues|wed|thur|thurs|fri|sat|sun'
    MONTHS_PATTERN = 'january|february|march|april|may|june|july|august|september|october|november|december|jan|feb|mar|apr|may|jun|jul|aug|sep|sept|oct|nov|dec'
    TIMEZONES_PATTERN = 'ACDT|ACST|ACT|ACWDT|ACWST|ADDT|ADMT|ADT|AEDT|AEST|AFT|AHDT|AHST|AKDT|AKST|AKTST|AKTT|ALMST|ALMT|AMST|AMT|ANAST|ANAT|ANT|APT|AQTST|AQTT|ARST|ART|ASHST|ASHT|AST|AWDT|AWST|AWT|AZOMT|AZOST|AZOT|AZST|AZT|BAKST|BAKT|BDST|BDT|BEAT|BEAUT|BIOT|BMT|BNT|BORT|BOST|BOT|BRST|BRT|BST|BTT|BURT|CANT|CAPT|CAST|CAT|CAWT|CCT|CDDT|CDT|CEDT|CEMT|CEST|CET|CGST|CGT|CHADT|CHAST|CHDT|CHOST|CHOT|CIST|CKHST|CKT|CLST|CLT|CMT|COST|COT|CPT|CST|CUT|CVST|CVT|CWT|CXT|ChST|DACT|DAVT|DDUT|DFT|DMT|DUSST|DUST|EASST|EAST|EAT|ECT|EDDT|EDT|EEDT|EEST|EET|EGST|EGT|EHDT|EMT|EPT|EST|ET|EWT|FET|FFMT|FJST|FJT|FKST|FKT|FMT|FNST|FNT|FORT|FRUST|FRUT|GALT|GAMT|GBGT|GEST|GET|GFT|GHST|GILT|GIT|GMT|GST|GYT|HAA|HAC|HADT|HAE|HAP|HAR|HAST|HAT|HAY|HDT|HKST|HKT|HLV|HMT|HNA|HNC|HNE|HNP|HNR|HNT|HNY|HOVST|HOVT|HST|ICT|IDDT|IDT|IHST|IMT|IOT|IRDT|IRKST|IRKT|IRST|ISST|IST|JAVT|JCST|JDT|JMT|JST|JWST|KART|KDT|KGST|KGT|KIZST|KIZT|KMT|KOST|KRAST|KRAT|KST|KUYST|KUYT|KWAT|LHDT|LHST|LINT|LKT|LMT|LMT|LMT|LMT|LRT|LST|MADMT|MADST|MADT|MAGST|MAGT|MALST|MALT|MART|MAWT|MDDT|MDST|MDT|MEST|MET|MHT|MIST|MIT|MMT|MOST|MOT|MPT|MSD|MSK|MSM|MST|MUST|MUT|MVT|MWT|MYT|NCST|NCT|NDDT|NDT|NEGT|NEST|NET|NFT|NMT|NOVST|NOVT|NPT|NRT|NST|NT|NUT|NWT|NZDT|NZMT|NZST|OMSST|OMST|ORAST|ORAT|PDDT|PDT|PEST|PET|PETST|PETT|PGT|PHOT|PHST|PHT|PKST|PKT|PLMT|PMDT|PMMT|PMST|PMT|PNT|PONT|PPMT|PPT|PST|PT|PWT|PYST|PYT|QMT|QYZST|QYZT|RET|RMT|ROTT|SAKST|SAKT|SAMT|SAST|SBT|SCT|SDMT|SDT|SET|SGT|SHEST|SHET|SJMT|SLT|SMT|SRET|SRT|SST|STAT|SVEST|SVET|SWAT|SYOT|TAHT|TASST|TAST|TBIST|TBIT|TBMT|TFT|THA|TJT|TKT|TLT|TMT|TOST|TOT|TRST|TRT|TSAT|TVT|ULAST|ULAT|URAST|URAT|UTC|UYHST|UYST|UYT|UZST|UZT|VET|VLAST|VLAT|VOLST|VOLT|VOST|VUST|VUT|WARST|WART|WAST|WAT|WDT|WEDT|WEMT|WEST|WET|WFT|WGST|WGT|WIB|WIT|WITA|WMT|WSDT|WSST|WST|WT|XJT|YAKST|YAKT|YAPT|YDDT|YDT|YEKST|YEKST|YEKT|YEKT|YERST|YERT|YPT|YST|YWT|zzz'
    ## explicit north american timezones that get replaced
    NA_TIMEZONES_PATTERN = 'pacific|eastern|mountain|central'
    ALL_TIMEZONES_PATTERN = TIMEZONES_PATTERN + '|' + NA_TIMEZONES_PATTERN
    DELIMITERS_PATTERN = r'[/\:\-\,\.\s\_\+\@]+'
    TIME_PERIOD_PATTERN = r'a\.m\.|am|p\.m\.|pm'
    ## can be in date strings but not recognized by dateutils
    EXTRA_TOKENS_PATTERN = 'due|by|on|during|standard|daylight|savings|time|date|dated|day|of|to|through|between|until|z|at|t'

    ## TODO: Get english numbers?
    ## http://www.rexegg.com/regex-trick-numbers-in-english.html

    RELATIVE_PATTERN = 'before|after|next|last|ago'
    TIME_SHORTHAND_PATTERN = 'noon|midnight|today|yesterday'
    UNIT_PATTERN = 'second|minute|hour|day|week|month|year'

    ## Time pattern is used independently, so specified here.
    TIME_PATTERN = r"""
    (?P<time>
        ## Captures in format XX:YY(:ZZ) (PM) (EST)
        (
            (?P<hours>\d{{1,2}})
            \:
            (?P<minutes>\d{{1,2}})
            (\:(?<seconds>\d{{1,2}}))?
            ([\.\,](?<microseconds>\d{{1,6}}))?
            \s*
            (?P<time_periods>{time_periods})?
            \s*
            (?P<timezones>{timezones})?
        )
        |
        ## Captures in format 11 AM (EST)
        ## Note with single digit capture requires time period
        (
            (?P<hours>\d{{1,2}})
            \s*
            (?P<time_periods>{time_periods})
            \s*
            (?P<timezones>{timezones})*
        )
    )
    """.format(
        time_periods=TIME_PERIOD_PATTERN,
        timezones=ALL_TIMEZONES_PATTERN
    )

    DATES_PATTERN = r"""
    (
        (
            {time}
            |
            ## Grab any digits
            (?P<digits_modifier>{digits_modifier})
            |
            (?P<digits>{digits})
            |
            (?P<days>{days})
            |
            (?P<months>{months})
            |
            ## Delimiters, ie Tuesday[,] July 18 or 6[/]17[/]2008
            ## as well as whitespace
            (?P<delimiters>{delimiters})
            |
            ## These tokens could be in phrases that dateutil does not yet recognize
            ## Some are US Centric
            (?P<extra_tokens>({extra_tokens})[\s\.$,;]+)
        ## We need at least three items to match for minimal datetime parsing
        ## ie 10pm
        ){{1,1}}
    )
    """

    ALL_GROUPS = ['time', 'digits_modifier', 'digits', 'days',
                  'months', 'delimiters', 'extra_tokens', 'timezones',
                  'time_periods', 'hours', 'minutes', 'seconds',
                  'microseconds']

    DATES_PATTERN = DATES_PATTERN.format(
        time=TIME_PATTERN,
        digits=DIGITS_PATTERN,
        digits_modifier=DIGITS_MODIFIER_PATTERN,
        days=DAYS_PATTERN,
        months=MONTHS_PATTERN,
        delimiters=DELIMITERS_PATTERN,
        extra_tokens=EXTRA_TOKENS_PATTERN
    )

    DATE_REGEX = re.compile(DATES_PATTERN, re.IGNORECASE | re.MULTILINE | re.UNICODE | re.DOTALL | re.VERBOSE)

    TIME_REGEX = re.compile(TIME_PATTERN, re.IGNORECASE | re.MULTILINE | re.UNICODE | re.DOTALL | re.VERBOSE)

    # split ranges
    RANGE_SPLIT_PATTERN = r'\Wto\W|\Wthrough\W'

    RANGE_SPLIT_REGEX = re.compile(RANGE_SPLIT_PATTERN,
                                   re.IGNORECASE | re.MULTILINE | re.UNICODE | re.DOTALL)

    ## These tokens can be in original text but dateutil
    ## won't handle them without modification
    REPLACEMENTS = {
        "standard": " ",
        "daylight": " ",
        "savings": " ",
        "time": " ",
        "date": " ",
        "by": " ",
        "due": " ",
        "on": " ",
        "to": " ",
    }

    TIMEZONE_REPLACEMENTS = {
        "pacific": "PST",
        "eastern": "EST",
        "mountain": "MST",
        "central": "CST",
    }

    ## Characters that can be removed from ends of matched strings
    STRIP_CHARS = ' \n\t:-.,_'

    def __init__(self, base_date=None):
        self.base_date = base_date

    def tokenize_string(self, text: str) -> List[Tuple[str, str, Dict[str, List[str]]]]:
        items = []  # type:List[Tuple[str, str, Dict[str, List[str]]]]

        last_index = 0

        for match in self.DATE_REGEX.finditer(text):
            match_str = match.group(0)
            indices = match.span(0)
            captures = match.capturesdict()
            for capt_key in captures:
                captures[capt_key] = [c.strip() for c in captures[capt_key]]
            group = self.get_token_group(captures)

            if indices[0] > last_index:
                items.append((text[last_index:indices[0]], '', {}))
            items.append((match_str, group, captures))
            last_index = indices[1]
        if last_index < len(text):
            items.append((text[last_index:len(text)], '', {}))
        return items

    def merge_tokens(self, tokens: List[Tuple[str, str]]) -> List[DateFragment]:
        MIN_MATCHES = 3
        fragments = []  # type:List[DateFragment]
        frag = DateFragment()

        start_char, total_chars = 0, 0

        for token in tokens:
            total_chars += len(token[0])

            tok_text, group, tok_capts = token[0], token[1], token[2]
            if not group:
                if frag.indices[1] > 0:
                    if frag.matches_count >= MIN_MATCHES:
                        fragments.append(frag)
                frag = DateFragment()
                start_char = total_chars
                continue

            frag.matches_count += 1
            if frag.indices[1] == 0:
                frag.indices = (start_char, total_chars)
            else:
                frag.indices = (frag.indices[0], total_chars)  # -1

            frag.match_str += tok_text

            for capt in tok_capts:
                if capt in frag.captures:
                    frag.captures[capt] += tok_capts[capt]
                else:
                    frag.captures[capt] = tok_capts[capt]

            start_char = total_chars

        if frag.matches_count >= MIN_MATCHES:  # frag.matches
            fragments.append(frag)

        for frag in fragments:
            for gr in self.ALL_GROUPS:
                if gr not in frag.captures:
                    frag.captures[gr] = []

        return fragments

    @staticmethod
    def get_token_group(captures: Dict[str, List[str]]) -> str:
        for gr in DateFinder.ALL_GROUPS:
            lst = captures.get(gr)
            if lst and len(lst) > 0:
                return gr
        return ''

    def extract_date_strings(self, text: str, strict=False):
        return self.extract_date_strings_inner(text, text_start=0, strict=strict)

    def extract_date_strings_inner(self, text: str, text_start: int = 0, strict=False):
        rng = self.split_date_range(text)
        if rng and len(rng) > 1:
            range_strings = []
            for range_str in rng:
                range_strings.extend(self.extract_date_strings_inner(range_str[0],
                                                                     range_str[1][0],
                                                                     strict=strict))
            for range_string in range_strings:
                yield range_string
            return

        tokens = self.tokenize_string(text)
        items = self.merge_tokens(tokens)

        for match in items:
            match_str = match.match_str
            indices = (match.indices[0] + text_start, match.indices[1] + text_start)

            ## Get individual group matches
            captures = match.captures
            #time = captures.get('time')
            digits = captures.get('digits')
            #digits_modifiers = captures.get('digits_modifiers')
            #days = captures.get('days')
            months = captures.get('months')
            #timezones = captures.get('timezones')
            delimiters = captures.get('delimiters')
            #time_periods = captures.get('time_periods')
            #extra_tokens = captures.get('extra_tokens')

            if delimiters and match_str.endswith(delimiters[-1]):
                captures['delimiters'] = captures['delimiters'][:-1]

            if strict:
                complete = False
                ## 12-05-2015
                if len(digits) == 3:
                    complete = True
                    ## 19 February 2013 year 09:10
                elif (len(months) == 1) and (len(digits) == 2):
                    complete = True

                if not complete:
                    continue

            ## sanitize date string
            ## replace unhelpful whitespace characters with single whitespace
            match_str = re.sub(r'[\n\t\s\xa0]+', ' ', match_str)
            match_str = match_str.strip(self.STRIP_CHARS)

            ## Save sanitized source string
            yield match_str, indices, captures

    @staticmethod
    def split_date_range(text: str) -> List[Tuple[str, Tuple[int, int]]]:
        st_matches = DateFinder.RANGE_SPLIT_REGEX.finditer(text)
        start = 0
        parts = []  # List[Tuple[str, Tuple[int, int]]]

        for match in st_matches:
            match_start = match.start()
            if match_start > start:
                parts.append((text[start:match_start], (start, match_start)))
            start = match.end()

        if start < len(text):
            parts.append((text[start:], (start, len(text))))

        return parts

    def find_dates(self, text, source=False, index=False, strict=False):

        for date_string, indices, captures in self.extract_date_strings(text, strict=strict):

            as_dt = self.parse_date_string(date_string, captures)
            if as_dt is None:
                ## Dateutil couldn't make heads or tails of it
                ## move on to next
                continue

            returnables = (as_dt,)
            if source:
                returnables = returnables + (date_string,)
            if index:
                returnables = returnables + (indices,)

            if len(returnables) == 1:
                returnables = returnables[0]
            yield returnables

    def parse_date_string(self, date_string, captures):
        # For well formatted string, we can already let dateutils parse them
        # otherwise self._find_and_replace method might corrupt them
        try:
            as_dt = parser.parse(date_string, default=self.base_date)
        except ValueError:
            # replace tokens that are problematic for dateutil
            date_string, tz_string = self._find_and_replace(date_string, captures)

            ## One last sweep after removing
            date_string = date_string.strip(self.STRIP_CHARS)
            ## Match strings must be at least 3 characters long
            ## < 3 tends to be garbage
            if len(date_string) < 3:
                return None

            try:
                debug_msg = 'Parsing {} with dateutil'.format(date_string)
                logger.debug(debug_msg)
                as_dt = parser.parse(date_string, default=self.base_date)
            except Exception as e:  # pylint: disable=broad-except
                logger.debug(e)
                as_dt = None
            if tz_string:
                as_dt = self._add_tzinfo(as_dt, tz_string)
        return as_dt

    def _add_tzinfo(self, datetime_obj, tz_string):
        """
        take a naive datetime and add dateutil.tz.tzinfo object

        :param datetime_obj: naive datetime object
        :return: datetime object with tzinfo
        """
        if datetime_obj is None:
            return None

        tzinfo_match = tz.gettz(tz_string)
        return datetime_obj.replace(tzinfo=tzinfo_match)

    def _find_and_replace(self, date_string, captures):
        """
        :warning: when multiple tz matches exist the last sorted capture will trump
        :param date_string:
        :return: date_string, tz_string
        """
        # add timezones to replace
        cloned_replacements = copy.copy(self.REPLACEMENTS)  # don't mutate
        for tz_string in captures.get('timezones', []):
            cloned_replacements.update({tz_string: ' '})

        date_string = date_string.lower()
        for key, replacement in cloned_replacements.items():
            # we really want to match all permutations of the key surrounded by whitespace chars except one
            # for example: consider the key = 'to'
            # 1. match 'to '
            # 2. match ' to'
            # 3. match ' to '
            # but never match r'(\s|)to(\s|)' which would make 'october' > 'ocber'
            date_string = re.sub(r'(^|\s)' + key + r'(\s|$)', replacement, date_string, flags=re.IGNORECASE)

        return date_string, self._pop_tz_string(sorted(captures.get('timezones', [])))

    def _pop_tz_string(self, list_of_timezones):
        try:
            tz_string = list_of_timezones.pop()
            # make sure it's not a timezone we
            # want replaced with better abbreviation
            return self.TIMEZONE_REPLACEMENTS.get(tz_string, tz_string)
        except IndexError:
            return ''


def find_dates(
        text,
        source=False,
        index=False,
        strict=False,
        base_date=None
):
    """
    Extract datetime strings from text

    :param text:
        A string that contains one or more natural language or literal
        datetime strings
    :type text: str|unicode
    :param source:
        Return the original string segment
    :type source: boolean
    :param index:
        Return the indices where the datetime string was located in text
    :type index: boolean
    :param strict:
        Only return datetimes with complete date information. For example:
        `July 2016` of `Monday` will not return datetimes.
        `May 16, 2015` will return datetimes.
    :type strict: boolean
    :param base_date:
        Set a default base datetime when parsing incomplete dates
    :type base_date: datetime

    :return: Returns a generator that produces :mod:`datetime.datetime` objects,
        or a tuple with the source text and index, if requested
    """
    date_finder = DateFinder(base_date=base_date)
    return date_finder.find_dates(text, source=source, index=index, strict=strict)



from enum import Enum

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class AnnotationLocatorType(Enum):
    RegexpBased = 1
    MlWordVectorBased = 2



import os

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


lexnlp_base_path = os.path.abspath(os.path.dirname(__file__) + '/../../../')

lexnlp_test_path = os.path.join(lexnlp_base_path, 'test_data/')



# pylint: disable=bare-except

# Imports

import types

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


def safe_failure(func):
    """
    return None on failure, either skip result if generator
    """
    def decorator(*args, **kwargs):
        raise_exc = not kwargs.pop('safe_failure', True)
        try:
            res = func(*args, **kwargs)
            if isinstance(res, types.GeneratorType):
                try:
                    yield from func(*args, **kwargs)
                except:
                    if raise_exc:
                        raise
        except:
            if raise_exc:
                raise
            return None

    return decorator



import codecs
import os
import shutil

import pandas
from typing import Tuple, List
from zipfile import ZipFile

from lexnlp.extract.common.annotations.definition_annotation import DefinitionAnnotation
from lexnlp.extract.ml.detector.detecting_settings import DetectingSettings
from lexnlp.extract.ml.detector.phrase_constructor import PhraseConstructorSettings, PhraseConstructorMethod
from lexnlp.extract.ml.en.definitions.definition_phrase_detector import DefinitionPhraseDetector
from lexnlp.extract.ml.en.definitions.definition_term_detector import DefinitionTermDetector

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class LayeredDefinitionDetector:
    def __init__(self):
        # let the prase be <agrees to serve the Company in such capacity during the
        # term of employment (the "Employment Period”).

        # ... model_definition will find <term of employment (the "Employment Period”)>
        self.model_definition = DefinitionPhraseDetector()
        # ... model_term will find
        self.model_term = DefinitionTermDetector()

        self.definition_join_sets = PhraseConstructorSettings(
            method=PhraseConstructorMethod.by_score, min_token_score=2, max_zeros=0)
        self.term_join_sets = PhraseConstructorSettings(
            method=PhraseConstructorMethod.by_class, strict=False)
        self.initialized = False

    def load_compressed(self, file_path: str):
        """
        Loads archive with two model pickle files (model_definition,
        model_term)
        """
        file_folder = os.path.dirname(file_path)
        temp_folder = os.path.join(file_folder, 'unpack_def_model_temp')
        try:
            shutil.rmtree(temp_folder)
        # pylint: disable=bare-except
        except:
            pass
        os.mkdir(temp_folder)
        
        with ZipFile(file_path) as z:
            z.extractall(temp_folder)

        model_files = [i for i in os.listdir(temp_folder) if i.endswith('.pickle')]
        for file_name in model_files:
            if file_name == 'definition.pickle':
                self.model_definition.load(os.path.join(temp_folder, file_name))
            elif file_name == 'term.pickle':
                self.model_term.load(os.path.join(temp_folder, file_name))
            else:
                raise RuntimeError(f'Found unknown file "{file_name.filename}" in packed model')
        shutil.rmtree(temp_folder)
        self.initialized = True

    def get_annotations(self, sentence: str) -> List[DefinitionAnnotation]:
        annotations = []  # type: List[DefinitionAnnotation]
        # we go from term to definition because term is a simplier object to locate
        terms = list(self.model_term.predict_text(
                sentence, join_settings=self.term_join_sets))
        if not terms:
            return annotations

        # find definitions around the terms
        # "mask" suggests the underlying model where the term is located
        feature_mask = [0] * len(sentence)
        for term in terms:
            for i in range(term[0], term[1]):
                feature_mask[i] = 1

        definitions = list(self.model_definition.predict_text(
                    sentence, feature_mask=feature_mask,
                    join_settings=self.definition_join_sets))

        # combine terms with surrounding definitions
        # measure distance between each tearm and each definition
        definition_distances = {}  # { term0: [0:d0, 1:d1, ...N:dN], term1: ... }
        for t_s, t_e in terms:
            distances = []
            for i in range(len(definitions)):
                df_s, df_e = definitions[i]
                if df_e <= t_s:
                    distance = t_s - df_e
                elif df_s >= t_e:
                    distance = df_s - t_e
                else:
                    if df_s > t_s:
                        distance = -(t_e - df_s + 1)
                    elif df_e < t_e:
                        distance = -(df_e - t_s + 1)
                    else:
                        distance = - (t_e - t_s + 1)
                distances.append((i, distance))
            distances.sort(key=lambda d: d[1])  # closest definitions go first
            definition_distances[(t_s, t_e)] = distances

        # get closest definitions for each term
        for term in terms:
            defs = definition_distances[term]
            df = defs[0] if defs else (term[0], term[1])
            def_start = min(df[0], term[0])
            def_end = max(df[1], term[1])

            term_phrase = sentence[term[0]: term[1]]
            ant = DefinitionAnnotation(
                coords=(def_start, def_end),
                name=term_phrase,
                text=sentence[def_start: def_end],
                locale='en')
            annotations.append(ant)

        # TODO: check if annotations overlap and cut overlapping parts
        return annotations

    def train_on_doccano_jsonl(self,
                               save_file_path: str,
                               exported_doc_path: str,
                               text_column_name: str = 'text',
                               labels_column_name: str = 'labels',
                               label_term: str = 'term',
                               label_definition: str = 'definition'):
        with codecs.open(exported_doc_path, 'rb', encoding='utf-8') as fr:
            df = pandas.read_json(fr, lines=True)
        df = df.drop(columns=['annotation_approver', 'id', 'meta'])

        definition_rows = []  # type: List[Tuple[str, List[Tuple[int, int]], List[int]]]
        term_rows = []  # type: List[Tuple[str, List[Tuple[int, int]]]]

        for _, row in df.iterrows():
            row_text = row[text_column_name]
            labels_definitions = []
            labels_terms = []
            for start, end, l_type in row[labels_column_name]:
                if l_type == label_definition:
                    labels_definitions.append((start, end))
                elif l_type == label_term:
                    labels_terms.append((start, end))

            # join definitions around terms because
            # definitions should (ideally) include terms, but due to
            # Doccano restrictions they !frame! terms. See:
            #   Member or sell or convey all or substantially all of its assets to any [d]Person that is not a
            #     Member (a "[/d][t]Merger Transaction[/t][d]")[/d] unless:

            definition_feature_mask, merged_def_labels = self.join_adjacent_definitions_labels(
                labels_definitions, labels_terms, row_text)

            definition_rows.append([row_text, merged_def_labels, definition_feature_mask])
            term_rows.append([row_text, labels_terms])

        df_terms = pandas.DataFrame(term_rows, columns=['sentence', 'labels'])
        df_definitions = pandas.DataFrame(definition_rows, columns=['sentence', 'labels', 'feature_mask'])
        self.train_on_formatted_data(df_definitions, df_terms, save_file_path)

    def train_on_formatted_data(self,
                                definition_frame: pandas.DataFrame,
                                term_frame: pandas.DataFrame,
                                save_file_path: str):
        """
        :param definition_frame: dataframe, [ (row_text, [(start, end), (start, end)...], feature_mask]
        :param term_frame: dataframe, [ (row_text, [(start, end), (start, end)...]]
        :param save_file_path: path to store zipped model files (as one file)
        """
        file_folder = os.path.dirname(save_file_path)
        temp_folder = os.path.join(file_folder, 'def_model_temp')
        try:
            shutil.rmtree(temp_folder)
        # pylint: disable=bare-except
        except:
            pass
        os.mkdir(temp_folder)

        file_terms = os.path.join(temp_folder, "terms")
        file_definitions = os.path.join(temp_folder, "definitions")

        self.model_term.train_and_save_on_dataframe(
            DetectingSettings(pre_window=1, post_window=1, use_spacy=False),
            term_frame,
            file_terms,
            compress=False)

        self.model_definition.train_and_save_on_dataframe(
            DetectingSettings(use_spacy=False),
            definition_frame,
            file_definitions,
            compress=False)

        with ZipFile(save_file_path, 'w') as zipObj2:
            zipObj2.write(file_terms, 'term.pickle')
            zipObj2.write(file_definitions, 'definition.pickle')
        try:
            shutil.rmtree(temp_folder)
        # pylint: disable=bare-except
        except:
            pass

    @staticmethod
    def join_adjacent_definitions_labels(labels_definitions, labels_terms, row_text):
        merged_def_labels = []
        definition_feature_mask = [0] * len(row_text)
        if not labels_definitions:
            return definition_feature_mask, merged_def_labels

        labels_definitions.sort(key=lambda l: l[0])
        ldef_end = -1
        for ld_start, ld_end in labels_definitions:
            continue_label = ldef_end >= 0 and abs(ld_start - ldef_end) < 2
            if not continue_label:
                merged_def_labels.append((ld_start, ld_end))
            else:
                merged_def_labels[-1] = (merged_def_labels[-1][0], ld_end)
            ldef_end = ld_end

            # continue current merged "definition" with "term"?
            for label_term_start, label_term_end in labels_terms:
                if abs(label_term_start - ldef_end) < 2:
                    ldef_end = label_term_end + 1
                    merged_def_labels[-1] = (merged_def_labels[-1][0], ldef_end)
                    for ci in range(label_term_start, label_term_end):
                        definition_feature_mask[ci] = 1
                    break

        return definition_feature_mask, merged_def_labels
